class InvalidNetworkError(Exception):
    """Exception raised when the network has an invalid configuration"""


class EngineNotFoundError(Exception):
    """
    Exception raised when no symbolic engine is found among the available ones.
    To see which are avaiable, see `sym_metanet.engines.get_available_engines`.
    """


class EngineNotFoundWarning(Warning):
    """
    Warning raised when no symbolic engine is found among the available ones.
    To see which are avaiable, see `sym_metanet.engines.get_available_engines`.
    """



from typing import TYPE_CHECKING, Collection, Dict, Literal, Optional, Tuple, Union

from sym_metanet.blocks.base import ElementWithVars
from sym_metanet.engines.core import EngineBase, get_current_engine
from sym_metanet.util.funcs import first
from sym_metanet.util.types import VarType

if TYPE_CHECKING:
    from sym_metanet.blocks.links import Link
    from sym_metanet.blocks.nodes import Node
    from sym_metanet.network import Network


class Origin(ElementWithVars[VarType]):
    """Ideal, state-less highway origin that conveys to the attached link as much flow
    as the flow in such link."""

    def init_vars(self, *args, **kwargs) -> None:
        """Initializes no variable in the ideal origin."""

    def step_dynamics(self, *args, **kwargs) -> Dict[str, VarType]:
        """No dynamics to steps in the ideal origin."""
        return {}

    def get_speed(self, net: "Network", **kwargs) -> VarType:
        """Computes the (upstream) speed induced by the ideal origin.

        Parameters
        ----------
        net : Network
            The network this destination belongs to.

        Returns
        -------
        variable
            The origin's upstream speed.
        """
        return self._get_exiting_link(net).states["v"][0]

    def get_flow(
        self, net: "Network", engine: Optional[EngineBase] = None, **kwargs
    ) -> VarType:
        """Computes the (upstream) flow induced by the ideal origin.

        Parameters
        ----------
        net : Network
            The network this destination belongs to.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        symbolic variable
            The origin's upstream flow.
        """
        return self._get_exiting_link(net).get_flow(engine)[0]

    def _get_exiting_link(self, net: "Network") -> "Link[VarType]":
        """Internal utility to fetch the link leaving this destination (can only be
        one)."""
        links_down: Collection[Tuple["Node", "Node", "Link[VarType]"]] = net.out_links(
            net.origins[self]  # type: ignore[index]
        )
        assert (
            len(links_down) == 1
        ), "Internal error. Only one link can leave an origin."
        return first(links_down)[-1]


class MeteredOnRamp(Origin[VarType]):
    """On-ramp where cars can queue up before being given access to the attached link.
    For reference, look at [1], in particular, Section 3.2.1 and Equations 3.5 and 3.6.

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    __slots__ = ("C", "flow_eq_type")
    _states = {"w"}
    _actions = {"r"}
    _disturbances = {"d"}

    def __init__(
        self,
        capacity: Union[VarType, float],
        flow_eq_type: Literal["in", "out"] = "out",
        name: Optional[str] = None,
    ) -> None:
        """Instantiates an on-ramp with the given capacity.

        Parameters
        ----------
        capacity : float or variable
            Capacity of the on-ramp, i.e., `C`.
        flow_eq_type : 'in' or 'out', optional
            Type of flow equation for the ramp. See
            `engine.origins.get_ramp_flow` for more details.
        name : str, optional
            Name of the on-ramp, by default None.
        """
        super().__init__(name)
        self.C = capacity
        self.flow_eq_type = flow_eq_type

    def init_vars(
        self,
        init_conditions: Optional[Dict[str, VarType]] = None,
        engine: Optional[EngineBase] = None,
    ) -> None:
        """Initializes
         - `w`: queue length (state)
         - `r`: ramp metering rate (control action)
         - `d`: demand (disturbance).

        Parameters
        ----------
        init_conditions : dict[str, variable], optional
            Provides name-variable tuples to initialize states, actions and disturbances
            with specific values. These values must be compatible with the symbolic
            engine in type and shape. If not provided, variables are initialized
            automatically.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.
        """
        if init_conditions is None:
            init_conditions = {}
        if engine is None:
            engine = get_current_engine()
        self.states: Dict[str, VarType] = {
            "w": init_conditions["w"]
            if "w" in init_conditions
            else engine.var(f"w_{self.name}")
        }
        self.actions: Dict[str, VarType] = {
            "r": init_conditions["r"]
            if "r" in init_conditions
            else engine.var(f"r_{self.name}")
        }
        self.disturbances: Dict[str, VarType] = {
            "d": init_conditions["d"]
            if "d" in init_conditions
            else engine.var(f"d_{self.name}")
        }

    def step_dynamics(
        self,
        net: "Network",
        T: Union[VarType, float],
        engine: Optional[EngineBase] = None,
        **kwargs,
    ) -> Dict[str, VarType]:
        """Steps the dynamics of this origin.

        Parameters
        ----------
        net : Network
            The network the origin belongs to.
        T : variable or float
            Sampling time.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        Dict[str, variable]
            A dict with the states of the origin (queue) at the next time step.
        """
        if engine is None:
            engine = get_current_engine()
        q = self.get_flow(net, T, engine, **kwargs)
        w_next = engine.origins.step_queue(
            self.states["w"], self.disturbances["d"], q, T
        )
        return {"w": w_next}

    def get_flow(  # type: ignore[override]
        self,
        net: "Network",
        T: Union[VarType, float],
        engine: Optional[EngineBase] = None,
        **kwargs,
    ) -> VarType:
        """Computes the (upstream) flow induced by the metered ramp.

        Parameters
        ----------
        net : Network
            The network this destination belongs to.
        T : variable or float
            Sampling time of the simulation.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        variable
            The origin's upstream flow.
        """
        if engine is None:
            engine = get_current_engine()
        link_down = self._get_exiting_link(net)
        return engine.origins.get_ramp_flow(
            self.disturbances["d"],
            self.states["w"],
            self.C,
            self.actions["r"],
            link_down.rho_max,
            link_down.states["rho"][0],
            link_down.rho_crit,
            T,
            self.flow_eq_type,
        )


class SimplifiedMeteredOnRamp(MeteredOnRamp[VarType]):
    """A simplified version of the vanilla on-ramp, where the flow of vehicles on the
    ramp is the direct control action (instead of controlling the metering rate that in
    turns dictates the car flow on the ramp).

    See `MeteredOnRamp` for the original version."""

    _actions = {"q"}

    def __init__(
        self,
        capacity: Union[VarType, float],
        flow_eq_type: Literal["limited", "unlimited"] = "limited",
        name: Optional[str] = None,
    ) -> None:
        """Instantiates a simplified on-ramp with the given capacity.

        Parameters
        ----------
        capacity : float or variable
            Capacity of the on-ramp, i.e., `C`.
        flow_eq_type : 'limited' or 'unlimited', optional
            Type of flow equation for the ramp. See
            `engine.origins.get_simplifiedramp_flow` for more details.
        name : str, optional
            Name of the on-ramp, by default None.
        """
        super().__init__(capacity, flow_eq_type, name)  # type: ignore[arg-type]

    def init_vars(
        self,
        init_conditions: Optional[Dict[str, VarType]] = None,
        engine: Optional[EngineBase] = None,
    ) -> None:
        """Initializes
         - `w`: queue length (state)
         - `q`: ramp flow (control action)
         - `d`: demand (disturbance).

        Parameters
        ----------
        init_conditions : dict[str, variable], optional
            Provides name-variable tuples to initialize states, actions and disturbances
            with specific values. These values must be compatible with the symbolic
            engine in type and shape. If not provided, variables are initialized
            automatically.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.
        """
        if init_conditions is None:
            init_conditions = {}
        if engine is None:
            engine = get_current_engine()
        self.states: Dict[str, VarType] = {
            "w": init_conditions["w"]
            if "w" in init_conditions
            else engine.var(f"w_{self.name}")
        }
        self.actions: Dict[str, VarType] = {
            "q": init_conditions["q"]
            if "q" in init_conditions
            else engine.var(f"q_{self.name}")
        }
        self.disturbances: Dict[str, VarType] = {
            "d": init_conditions["d"]
            if "d" in init_conditions
            else engine.var(f"d_{self.name}")
        }

    def get_flow(  # type: ignore[override]
        self,
        net: "Network",
        T: Union[VarType, float],
        engine: Optional[EngineBase] = None,
        *args,
        **kwargs,
    ) -> VarType:
        """Computes the (upstream) flow induced by the simple-metered ramp.

        Returns
        -------
        variable
            The origin's upstream flow.
        """
        if engine is None:
            engine = get_current_engine()
        link_down = self._get_exiting_link(net)
        return engine.origins.get_simplifiedramp_flow(
            self.actions["q"],
            self.disturbances["d"],
            self.states["w"],
            self.C,
            link_down.rho_max,
            link_down.states["rho"][0],
            link_down.rho_crit,
            T,
            self.flow_eq_type,  # type: ignore[arg-type]
        )



from itertools import product
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
)

import casadi as cs

from sym_metanet.engines.core import (
    DestinationsEngineBase,
    EngineBase,
    LinksEngineBase,
    NodesEngineBase,
    OriginsEngineBase,
)

if TYPE_CHECKING:
    from sym_metanet.blocks.links import Link
    from sym_metanet.network import Network


VarType = TypeVar("VarType", cs.SX, cs.MX)


class NodesEngine(NodesEngineBase, Generic[VarType]):
    """CasADi implementation of `sym_metanet.engines.core.NodesEngineBase`."""

    @staticmethod
    def get_upstream_flow(
        q_lasts: VarType,
        beta: VarType,
        betas: VarType,
        q_orig: Optional[VarType] = None,
    ) -> VarType:
        Q = cs.sum1(q_lasts)
        if q_orig is not None:
            Q += q_orig
        return (beta / cs.sum1(betas)) * Q

    @staticmethod
    def get_upstream_speed(q_lasts: VarType, v_lasts: VarType) -> VarType:
        return cs.sum1(v_lasts * q_lasts) / cs.sum1(q_lasts)

    @staticmethod
    def get_downstream_density(rho_firsts: VarType) -> VarType:
        return cs.sum1(rho_firsts**2) / cs.sum1(rho_firsts)


class LinksEngine(LinksEngineBase, Generic[VarType]):
    """CasADi implementation of `sym_metanet.engines.core.LinksEngineBase`."""

    @staticmethod
    def get_flow(rho: VarType, v: VarType, lanes: VarType) -> VarType:
        return rho * v * lanes

    @staticmethod
    def step_density(
        rho: VarType, q: VarType, q_up: VarType, lanes: VarType, L: VarType, T: VarType
    ) -> VarType:
        return rho + (T / lanes / L) * (q_up - q)

    @staticmethod
    def step_speed(
        v: VarType,
        v_up: VarType,
        rho: VarType,
        rho_down: VarType,
        Veq: VarType,
        lanes: VarType,
        L: VarType,
        tau: VarType,
        eta: VarType,
        kappa: VarType,
        T: VarType,
        q_ramp: Optional[VarType] = None,
        delta: Optional[VarType] = None,
        lanes_drop: Optional[VarType] = None,
        phi: Optional[VarType] = None,
        rho_crit: Optional[VarType] = None,
    ) -> VarType:
        relaxation = (T / tau) * (Veq - v)
        convection = T * v / L * (v_up - v)
        anticipation = (eta * T / tau) * (rho_down - rho) / (L * (rho + kappa))
        v_next = v + relaxation + convection - anticipation
        if q_ramp is not None and delta is not None:
            v_next[0] -= (delta * T * q_ramp * v[0]) / (L * lanes * (rho[0] + kappa))
        if lanes_drop is not None and phi is not None and rho_crit is not None:
            v_next[-1] -= (phi * T * lanes_drop * rho[-1] * v[-1] ** 2) / (
                L * lanes * rho_crit
            )
        return v_next

    @staticmethod
    def Veq(rho: VarType, v_free: VarType, rho_crit: VarType, a: VarType) -> VarType:
        return v_free * cs.exp((-1 / a) * cs.power(rho / rho_crit, a))


class OriginsEngine(OriginsEngineBase, Generic[VarType]):
    """CasADi implementation of `sym_metanet.engines.core.OriginsEngineBase`."""

    @staticmethod
    def step_queue(w: VarType, d: VarType, q: VarType, T: VarType) -> VarType:
        return w + T * (d - q)

    @staticmethod
    def get_ramp_flow(
        d: VarType,
        w: VarType,
        C: VarType,
        r: VarType,
        rho_max: VarType,
        rho_first: VarType,
        rho_crit: VarType,
        T: VarType,
        type: Literal["in", "out"] = "out",
    ) -> VarType:
        term1 = d + w / T
        term3 = (rho_max - rho_first) / (rho_max - rho_crit)
        if type == "in":
            return cs.fmin(term1, C * cs.fmin(r, term3))
        return r * cs.fmin(term1, C * cs.fmin(1, term3))

    @staticmethod
    def get_simplifiedramp_flow(
        qdes: VarType,
        d: VarType = None,
        w: VarType = None,
        C: VarType = None,
        rho_max: VarType = None,
        rho_first: VarType = None,
        rho_crit: VarType = None,
        T: VarType = None,
        type: Literal["limited", "unlimited"] = "limited",
    ) -> VarType:
        if type == "unlimited":
            return qdes
        term2 = d + w / T
        term3 = C * cs.fmin(1, (rho_max - rho_first) / (rho_max - rho_crit))
        return cs.fmin(qdes, cs.fmin(term2, term3))


class DestinationsEngine(DestinationsEngineBase, Generic[VarType]):
    """CasADi implementation of `sym_metanet.engines.core.DestinationsEngineBase`."""

    @staticmethod
    def get_congested_downstream_density(
        rho_last: VarType, rho_destination: VarType, rho_crit: VarType
    ) -> VarType:
        return cs.fmax(cs.fmin(rho_last, rho_crit), rho_destination)


class Engine(EngineBase, Generic[VarType]):
    """Symbolic engine implemented with the CasADi framework"""

    def __init__(self, sym_type: Literal["SX", "MX"] = "SX") -> None:
        """Instantiates a CasADi engine.

        Parameters
        ----------
        sym_type : {'SX', 'MX'}, optional
            A string that tells the engine with type of symbolic variables to use. Must
            be either `'SX'` or `'MX'`, at which point the engine employes `casadi.SX`
            or `casadi.MX` variables, respectively. By default, `'SX'` is used.

        Raises
        ------
        AttributeError
            Raises if `sym_type` is not valid.
        """
        super().__init__()
        self.sym_type: Union[Type[cs.SX], Type[cs.MX]] = getattr(cs, sym_type)

    @property
    def nodes(self) -> Type[NodesEngine[VarType]]:
        return NodesEngine[VarType]

    @property
    def links(self) -> Type[LinksEngine[VarType]]:
        return LinksEngine[VarType]

    @property
    def origins(self) -> Type[OriginsEngine[VarType]]:
        return OriginsEngine[VarType]

    @property
    def destinations(self) -> Type[DestinationsEngine[VarType]]:
        return DestinationsEngine[VarType]

    def var(self, name: str, n: int = 1, *args, **kwargs) -> VarType:
        return self.sym_type.sym(name, n, 1)

    def vcat(self, *arrays):
        return cs.vertcat(*arrays)

    def to_function(  # type: ignore[override]
        self,
        net: "Network",
        compact: int = 0,
        more_out: bool = False,
        force_positive_speed: bool = True,
        force_positive_density: bool = False,
        force_positive_queue: bool = False,
        parameters: Optional[Dict[str, VarType]] = None,
        **other_parameters,
    ) -> cs.Function:
        """Converts the network's dynamics to a CasADi Function.

        Parameters
        ----------
        net : Network
            The network whose dynamics must be translated into a function.
        compact : int, optional
            The compactness of input and output arguments. The levels are

            - <= 0: no aggregation of arguments, i.e., the function keeps states, action
            or disturbances for each element separate.

            - == 1: some aggregation, i.e., same variable types are clumped together.

            -  > 1: most aggregation, i.e., states, action and disturbances are
            aggregated in a single vector each.

        more_out : bool, optional
            Includes flows of links and origins in the output. By default `False`.
        force_positive_speed : bool, optional
            If `True`, the links speeds at the next time step are forced to be positive
            as `v+ = max(0, v+)`. METANET is in fact known to sometime yield negative
            speeds, which are infeasible.
        force_positive_density : bool, optional
            Same as `force_positive_speed`, but for densities. By default, `False`.
        force_positive_queue : bool, optional
            Same as `force_positive_speed`, but for queues. By default, `False`.
        parameters : dict[str, casadi.SX or MX], optional
            Symbolic network parameters to be included in the function, by default None.
        **other_parameters
            Other parameters (numerical or symbolical) required during the computations,
            e.g., sampling time T is usually required.

        Returns
        -------
        cs.Function
            The CasADi Function representing the network's dynamics.

        Raises
        ------
        RuntimeError
            Raises if variables have not yet been initialized; or if the dynamics have
            not been stepped yet, so no state at the next time instant is found.
        """
        for el, group in product(
            net.elements, ["_states", "_actions", "_disturbances"]
        ):
            if any(getattr(el, group)) and not getattr(el, f"has{group}"):
                raise RuntimeError(
                    f"Found no {group[1:-1]} in {el.name}; perhaps variables "
                    "have not been initialized via `net.init_vars`?"
                )
            if any(el._states) and not el.has_next_states:
                raise RuntimeError(
                    f"Found no next state in {el.name}; perhaps dynamics have "
                    "not been stepped via `net.step`?"
                )

        if parameters is None:
            parameters = {}

        # process inputs
        x = {el: _filter_vars(vars) for el, vars in net.states.items()}
        u = {el: _filter_vars(vars) for el, vars in net.actions.items()}
        d = {el: _filter_vars(vars) for el, vars in net.disturbances.items()}

        # gather inputs
        names_in, args_in = [], []
        if compact <= 0:
            for vars_in in (x, u, d):
                for el, vars in vars_in.items():  # type: ignore[attr-defined]
                    for varname, var in vars.items():
                        names_in.append(f"{varname}_{el.name}")
                        args_in.append(var)
        else:
            states: Dict[str, List[VarType]] = {}
            actions: Dict[str, List[VarType]] = {}
            disturbances: Dict[str, List[VarType]] = {}
            for vars_in, group in [(x, states), (u, actions), (d, disturbances)]:
                for el, vars in vars_in.items():  # type: ignore[attr-defined]
                    for varname, var in vars.items():
                        if varname in group:
                            group[varname].append(var)
                        else:
                            group[varname] = [var]

            for group in (states, actions, disturbances):
                for varname, vars in group.items():
                    group[varname] = cs.vertcat(*vars)

            if compact == 1:
                names_in = (
                    list(states.keys())
                    + list(actions.keys())
                    + list(disturbances.keys())
                )
                args_in = (
                    list(states.values())
                    + list(actions.values())
                    + list(disturbances.values())
                )
            else:
                names_in = ["x", "u", "d"]
                args_in = [
                    cs.vertcat(*states.values()),
                    cs.vertcat(*actions.values()),
                    cs.vertcat(*disturbances.values()),
                ]

        # process outputs
        x_next = {
            el: _filter_vars(vars, independent=False)
            for el, vars in net.next_states.items()
        }
        if force_positive_speed or force_positive_density or force_positive_queue:
            for vars in x_next.values():
                if force_positive_speed and "v" in vars:
                    vars["v"] = cs.fmax(0, vars["v"])
                if force_positive_density and "rho" in vars:
                    vars["rho"] = cs.fmax(0, vars["rho"])
                if force_positive_queue and "w" in vars:
                    vars["w"] = cs.fmax(0, vars["w"])

        # gather outputs
        names_out, args_out = [], []
        if compact <= 0:
            for el, vars in x_next.items():
                for varname, var in vars.items():
                    names_out.append(f"{varname}_{el.name}+")
                    args_out.append(var)
        else:
            next_states: Dict = {}
            for vars in x_next.values():
                for varname, var in vars.items():
                    varname += "+"
                    if varname in next_states:
                        next_states[varname].append(var)
                    else:
                        next_states[varname] = [var]

            for varname, vars in next_states.items():
                next_states[varname] = cs.vertcat(*vars)

            if compact == 1:
                names_out = list(next_states.keys())
                args_out = list(next_states.values())
            else:
                names_out = ["x+"]
                args_out = [cs.vertcat(*next_states.values())]

        # add link and origin flows (q, q_o) to output
        if more_out:
            names_link: List[str] = []
            flows_link: List[VarType] = []
            names_origins, flows_origins = [], []
            link: "Link[VarType]"
            for _, _, link in net.links:
                names_link.append(f"q_{link.name}")
                flows_link.append(link.get_flow(self))
            for origin in net.origins:
                names_origins.append(f"q_o_{origin.name}")
                flows_origins.append(
                    origin.get_flow(net, engine=self, **parameters, **other_parameters)
                )

            if compact > 0:
                names_link = ["q"]
                flows_link = [cs.vertcat(*flows_link)]
                names_origins = ["q_o"]
                flows_origins = [cs.vertcat(*flows_origins)]
            if compact > 1:
                names_link = ["q"]
                flows_link = [cs.vertcat(flows_link[0], flows_origins[0])]
                names_origins, flows_origins = [], []
            names_out += names_link + names_origins
            args_out += flows_link + flows_origins

        # add parameters
        if parameters:
            if compact <= 0:
                names_in.extend(parameters.keys())
                args_in.extend(parameters.values())
            else:
                names_in.append("p")
                args_in.append(cs.vertcat(*parameters.values()))

        return cs.Function("F", args_in, args_out, names_in, names_out)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(casadi)"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(casadi, type={self.sym_type.__name__})"


def _filter_vars(
    vars: Dict[str, Union[VarType, Any]],
    symbolic: bool = True,
    independent: bool = True,
) -> Dict[str, VarType]:
    """Internal utility to filter out symbols that are either only symbolic
    and/or independent (and thus can be inputs to `casadi.Function`)."""

    def is_ok(var: Union[VarType, Any]) -> bool:
        # sourcery skip: assign-if-exp, reintroduce-else
        if not symbolic:
            return True
        if isinstance(var, cs.SX):
            if independent:
                return all(var[i].n_dep() == 0 for i in range(var.shape[0]))
            return True
        if isinstance(var, cs.MX):
            if independent:
                return var.n_dep() == 0
            return True
        return False

    return {name: var for name, var in vars.items() if is_ok(var)}



from typing import Callable, Literal, Optional, Type, Union

import numpy as np

from sym_metanet.engines.core import (
    DestinationsEngineBase,
    EngineBase,
    LinksEngineBase,
    NodesEngineBase,
    OriginsEngineBase,
)


class NodesEngine(NodesEngineBase):
    """NumPy implementation of `sym_metanet.engines.core.NodesEngineBase`."""

    @staticmethod
    def get_upstream_flow(
        q_lasts: np.ndarray,
        beta: np.ndarray,
        betas: np.ndarray,
        q_orig: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        Q = np.sum(q_lasts, 0)
        if q_orig is not None:
            Q += q_orig
        return (beta / np.sum(betas, 0)) * Q

    @staticmethod
    def get_upstream_speed(q_lasts: np.ndarray, v_lasts: np.ndarray) -> np.ndarray:
        return np.sum(v_lasts * q_lasts, 0) / np.sum(q_lasts, 0)

    @staticmethod
    def get_downstream_density(rho_firsts: np.ndarray) -> np.ndarray:
        return np.sum(np.square(rho_firsts), 0) / np.sum(rho_firsts, 0)


class LinksEngine(LinksEngineBase):
    """NumPy implementation of `sym_metanet.engines.core.LinksEngineBase`."""

    @staticmethod
    def get_flow(rho: np.ndarray, v: np.ndarray, lanes: np.ndarray) -> np.ndarray:
        return rho * v * lanes

    @staticmethod
    def step_density(
        rho: np.ndarray,
        q: np.ndarray,
        q_up: np.ndarray,
        lanes: np.ndarray,
        L: np.ndarray,
        T: np.ndarray,
    ) -> np.ndarray:
        return rho + (T / lanes / L) * (q_up - q)

    @staticmethod
    def step_speed(
        v: np.ndarray,
        v_up: np.ndarray,
        rho: np.ndarray,
        rho_down: np.ndarray,
        Veq: np.ndarray,
        lanes: int,
        L: int,
        tau: float,
        eta: float,
        kappa: float,
        T: float,
        q_ramp: Optional[np.ndarray] = None,
        delta: Optional[float] = None,
        lanes_drop: Optional[int] = None,
        phi: Optional[float] = None,
        rho_crit: Optional[float] = None,
    ) -> np.ndarray:
        relaxation = (T / tau) * (Veq - v)
        convection = T * v / L * (v_up - v)
        anticipation = (eta * T / tau) * (rho_down - rho) / (L * (rho + kappa))
        v_next = v + relaxation + convection - anticipation
        if q_ramp is not None and delta is not None:
            v_next[0] -= (delta * T * q_ramp * v[0]) / (L * lanes * (rho[0] + kappa))
        if lanes_drop is not None and phi is not None and rho_crit is not None:
            v_next[-1] -= (phi * T * lanes_drop * rho[-1] * v[-1] ** 2) / (
                L * lanes * rho_crit
            )
        return v_next

    @staticmethod
    def Veq(
        rho: np.ndarray, v_free: np.ndarray, rho_crit: np.ndarray, a: np.ndarray
    ) -> np.ndarray:
        return v_free * np.exp((-1 / a) * np.power(rho / rho_crit, a))


class OriginsEngine(OriginsEngineBase):
    """
    NumPy implementation of `sym_metanet.engines.core.OriginsEngineBase`.
    """

    @staticmethod
    def step_queue(
        w: np.ndarray, d: np.ndarray, q: np.ndarray, T: np.ndarray
    ) -> np.ndarray:
        return w + T * (d - q)

    @staticmethod
    def get_ramp_flow(
        d: np.ndarray,
        w: np.ndarray,
        C: np.ndarray,
        r: np.ndarray,
        rho_max: np.ndarray,
        rho_first: np.ndarray,
        rho_crit: np.ndarray,
        T: np.ndarray,
        type: Literal["in", "out"] = "out",
    ) -> np.ndarray:
        term1 = d + w / T
        term3 = (rho_max - rho_first) / (rho_max - rho_crit)
        if type == "in":
            return np.minimum(term1, C * np.minimum(r, term3))
        return r * np.minimum(term1, C * np.minimum(1, term3))

    @staticmethod
    def get_simplifiedramp_flow(
        qdes: np.ndarray,
        d: Optional[np.ndarray] = None,
        w: Optional[np.ndarray] = None,
        C: Optional[np.ndarray] = None,
        rho_max: Optional[np.ndarray] = None,
        rho_first: Optional[np.ndarray] = None,
        rho_crit: Optional[np.ndarray] = None,
        T: Optional[np.ndarray] = None,
        type: Literal["limited", "unlimited"] = "limited",
    ) -> np.ndarray:
        if type == "unlimited":
            return qdes
        term2 = d + w / T  # type: ignore[operator]
        term3 = C * np.minimum(
            1, (rho_max - rho_first) / (rho_max - rho_crit)  # type: ignore[operator]
        )
        return np.minimum(qdes, np.minimum(term2, term3))


class DestinationsEngine(DestinationsEngineBase):
    """
    NumPy implementation of `sym_metanet.engines.core.DestinationsEngineBase`.
    """

    @staticmethod
    def get_congested_downstream_density(
        rho_last: np.ndarray, rho_destination: np.ndarray, rho_crit: np.ndarray
    ) -> np.ndarray:
        return np.maximum(np.minimum(rho_last, rho_crit), rho_destination)


class Engine(EngineBase):
    """Symbolic engine implemented with NumPy."""

    def __init__(self, var_type: Union[str, np.ndarray] = "empty") -> None:
        """Instantiates a NumPy engine.

        Parameters
        ----------
        var_type : str or array_like, optional
            How numerical values inside each variable should be initialized

             - 'empty': variable vectors are initialized as empty

             - 'rand': variable vectors are initialized randomly (uniform)

             - 'randn': variable vectors are initialized randomly (natural)

            or initialized with the given array value. By default, 'empty' is
            selected.

        Raises
        ------
        ValueError
            Raises if the specified type of variable initialization is invalid.
        """
        super().__init__()
        self.var_type = var_type

    @property
    def var_type(self) -> Union[str, np.ndarray]:
        """Gets the variable initialization strategy."""
        return self._var_type

    @var_type.setter
    def var_type(self, val: Union[str, np.ndarray]) -> None:
        """Sets the variable initialization strategy.

        Parameters
        ----------
        var_type : str or array_like, optional
            How numerical values inside each variable should be initialized

             - 'empty': variable vectors are initialized as empty

             - 'rand': variable vectors are initialized randomly (uniform)

             - 'randn': variable vectors are initialized randomly (natural)

            or initialized with the given array value.

        Raises
        ------
        ValueError
            Raises if the specified type of variable initialization is invalid.
        """
        if isinstance(val, str):
            if val == "empty":

                def gen(n):
                    return np.empty((n,), float)

            elif val == "rand":

                def gen(n):
                    return np.random.rand(n)

            elif val == "randn":

                def gen(n):
                    return np.random.randn(n)

            else:
                raise ValueError("Invalid variable initialization method.")
        else:

            def gen(n):
                return np.full((n,), val)

        self._var_gen: Callable[[int], np.ndarray] = gen
        self._var_type = val

    @property
    def nodes(self) -> Type[NodesEngine]:
        return NodesEngine

    @property
    def links(self) -> Type[LinksEngine]:
        return LinksEngine

    @property
    def origins(self) -> Type[OriginsEngine]:
        return OriginsEngine

    @property
    def destinations(self) -> Type[DestinationsEngine]:
        return DestinationsEngine

    def var(self, name: str, n: int = 1, *args, **kwargs) -> np.ndarray:
        return self._var_gen(n)

    def vcat(self, *arrays: np.ndarray):
        return np.hstack(arrays)

    def to_function(self, *args, **kwargs) -> Callable:
        raise NotImplementedError("`to_function` not available for NumPy engine.")

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(numpy)"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(numpy)"



from typing import Protocol, TypeVar, Union


class SupportsItem(Protocol):
    """An ABC which is indexable."""

    __slots__ = ()

    def __getitem__(self, i):
        pass

    def __setitem__(self, i, item):
        pass


class SupportsAlgebraOps(Protocol):
    """An ABC that supports simple algebraic computations like +, -, etc."""

    __slots__ = ()

    def __add__(
        self, x: Union["SupportsAlgebraOps", int, float]
    ) -> "SupportsAlgebraOps":
        pass

    def __sub__(
        self, x: Union["SupportsAlgebraOps", int, float]
    ) -> "SupportsAlgebraOps":
        pass


class Variable(SupportsAlgebraOps, SupportsItem):
    "Variable that can also be numerical or symbolic, depending on the engine. It must"
    "be indexable as an array, in case of vector quantities need to be represented."


VarType = TypeVar("VarType", bound=Variable)



from typing import TYPE_CHECKING, Collection, Dict, Optional, Tuple, Union

from sym_metanet.blocks.base import ElementWithVars
from sym_metanet.blocks.origins import MeteredOnRamp
from sym_metanet.engines.core import EngineBase, get_current_engine
from sym_metanet.util.funcs import first
from sym_metanet.util.types import VarType

if TYPE_CHECKING:
    from sym_metanet.blocks.nodes import Node
    from sym_metanet.network import Network


class Link(ElementWithVars[VarType]):
    """Highway link between two nodes [1, Section 3.2.1]. Links represent stretch of
    highway with similar traffic characteristics and no road changes (e.g., same number
    of lanes and maximum speed).

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    __slots__ = ("N", "lam", "L", "rho_max", "rho_crit", "v_free", "a", "turnrate")
    _states = {"rho", "v"}

    def __init__(
        self,
        nb_segments: int,
        lanes: Union[VarType, int],
        length: Union[VarType, float],
        maximum_density: Union[VarType, float],
        critical_density: Union[VarType, float],
        free_flow_velocity: Union[VarType, float],
        a: Union[VarType, float],
        turnrate: Union[VarType, float] = 1.0,
        name: Optional[str] = None,
    ) -> None:
        """Creates an instance of a METANET link.

        Parameters
        ----------
        nb_segments : int
            Number of segments in this highway link, i.e., `N`.
        lanes : int or variable
            Number of lanes in each segment, i.e., `lam`.
        lengths : float or variable
            Length of each segment in the link, i.e., `L`.
        maximum density : float or variable
            Maximum density that the link can withstand, i.e., `rho_max`.
        critical_densities : float or variable
            Critical density at which the traffic flow is maximal, i.e., `rho_crit`.
        free_flow_velocities : float or variable
            Average speed of cars when traffic is freely flowing, i.e., `v_free`.
        a : float or variable
            Model parameter in the computations of the equivalent speed [1, Equation
            3.4].
        turnrate : float or variable, optional
            Fraction of the total flow that enters this link via the upstream node. Only
            relevant if multiple exiting links are attached to the same node, in order
            to split the flow according to these rates. Needs not be normalized. By
            default, all links have equal rates.
        name : str, optional
            Name of this link, by default `None`.

        References
        ----------
        [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
            measures", Netherlands TRAIL Research School.
        """
        super().__init__(name)
        self.N = nb_segments
        self.lam = lanes
        self.L = length
        self.rho_max = maximum_density
        self.rho_crit = critical_density
        self.v_free = free_flow_velocity
        self.a = a
        self.turnrate = turnrate

    def init_vars(
        self,
        init_conditions: Optional[Dict[str, VarType]] = None,
        engine: Optional[EngineBase] = None,
    ) -> None:
        """For each segment in the link, initializes
         - `rho`: densities (state)
         - `v`: speeds (state).

        Parameters
        ----------
        init_conditions : dict[str, variable], optional
            Provides name-variable tuples to initialize states, actions and disturbances
            with specific values. These values must be compatible with the symbolic
            engine in type and shape. If not provided, variables are initialized
            automatically.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.
        """
        if init_conditions is None:
            init_conditions = {}
        if engine is None:
            engine = get_current_engine()
        self.states: Dict[str, VarType] = {
            name: (
                init_conditions[name]
                if name in init_conditions
                else engine.var(f"{name}_{self.name}", self.N)
            )
            for name in ("rho", "v")
        }

    def get_flow(self, engine: Optional[EngineBase] = None, **kwargs) -> VarType:
        """Gets the flow in this link's segments.

        Parameters
        ----------
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        variable
            The flow in this link.
        """
        if engine is None:
            engine = get_current_engine()
        return engine.links.get_flow(self.states["rho"], self.states["v"], self.lam)

    def step_dynamics(
        self,
        net: "Network",
        tau: Union[VarType, float],
        eta: Union[VarType, float],
        kappa: Union[VarType, float],
        T: Union[VarType, float],
        delta: Union[None, VarType, float] = None,
        phi: Union[None, VarType, float] = None,
        engine: Optional[EngineBase] = None,
        **kwargs,
    ) -> Dict[str, VarType]:
        """Steps the dynamics of this link.

        Parameters
        ----------
        net : Network
            The network the link belongs to.
        tau : float or variable
            Model parameter for the speed relaxation term.
        eta : float or variable
            Model parameter for the speed anticipation term.
        kappa : float or variable
            Model parameter for the speed anticipation term.
        T : float or variable
            Sampling time.
        delta : float or variable, optional
            Model parameter for merging phenomenum. By default, not considered.
        phi : float or variable, optional
            Model parameter for lane drop phenomenum. By defaul, not considered.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        Dict[str, variable]
            A dict with the states of the link (speeds and densities) at the next time
            step.
        """
        if engine is None:
            engine = get_current_engine()

        node_up, node_down = net.nodes_by_link[self]  # type: ignore[index]
        rho = self.states["rho"]
        v = self.states["v"]
        q = self.get_flow(engine)

        # get upstream flow and speed, and downstream density
        v0, q0 = node_up.get_upstream_speed_and_flow(net, self, engine, T=T)
        rhoN_1 = node_down.get_downstream_density(net, engine)
        if self.N > 1:
            q_up = engine.vcat(q0, q[:-1])
            v_up = engine.vcat(v0, v[:-1])
            rho_down = engine.vcat(rho[1:], rhoN_1)
        else:
            q_up = q0
            v_up = v0
            rho_down = rhoN_1

        # check for ramp merging in this link's upstream node with other
        # entering links.
        q_ramp = None
        if (
            delta is not None
            and node_up in net.origins_by_node
            and any(net.in_links(node_up))
        ):
            origin = net.origins_by_node[node_up]
            if isinstance(origin, MeteredOnRamp):
                q_ramp = origin.get_flow(net, T, engine)

        # check for lane drops in the next link (only if one link downstream)
        lanes_drop = None
        if phi is not None:
            links_down: Collection[
                Tuple["Node", "Node", "Link[VarType]"]
            ] = net.out_links(node_down)
            if len(links_down) == 1:
                link_down = first(links_down)[-1]
                lanes_drop = self.lam - link_down.lam  # type: ignore[operator]
            if lanes_drop == 0:
                lanes_drop = None

        # step densities
        rho_next = engine.links.step_density(rho, q, q_up, self.lam, self.L, T)

        # step speeds
        Veq = engine.links.Veq(rho, self.v_free, self.rho_crit, self.a)
        v_next = engine.links.step_speed(
            v,
            v_up,
            rho,
            rho_down,
            Veq,
            self.lam,
            self.L,
            tau,
            eta,
            kappa,
            T,
            q_ramp,
            delta,
            lanes_drop,
            phi,
            self.rho_crit,
        )
        return {"rho": rho_next, "v": v_next}



from functools import cached_property
from itertools import chain, product
from typing import Dict, Iterable, List, Optional, Tuple, Union

import networkx as nx

from sym_metanet.blocks.base import ElementBase, ElementWithVars
from sym_metanet.blocks.destinations import Destination
from sym_metanet.blocks.links import Link
from sym_metanet.blocks.nodes import Node
from sym_metanet.blocks.origins import MeteredOnRamp, Origin
from sym_metanet.engines.core import EngineBase
from sym_metanet.errors import InvalidNetworkError
from sym_metanet.util.funcs import invalidate_cache
from sym_metanet.util.types import VarType
from sym_metanet.views import (
    DESTINATIONENTRY,
    LINKENTRY,
    ORIGINENTRY,
    InLinkViewWrapper,
    OutLinkViewWrapper,
)


class Network(ElementBase):
    """Highway network."""

    def __init__(self, name: Optional[str] = None):
        """Instantiates an highway network.

        Parameters
        ----------
        name : str, optional
            Name of the network, by default `None`.
        """
        super().__init__(name)
        self._graph = nx.DiGraph(name=name)

    @property
    def G(self) -> nx.DiGraph:
        """Returns the underlying `networkx`'s graph of the highway."""
        return self._graph

    @property
    def graph(self) -> nx.DiGraph:
        """Returns the underlying `networkx`'s graph of the highway."""
        return self._graph

    @property
    def asgraph(self) -> nx.DiGraph:
        """Returns the underlying `networkx`'s graph of the highway."""
        return self._graph

    @property
    def nodes(self) -> nx.classes.reportviews.NodeView:
        """Returns a view on the nodes of the network."""
        return self._graph.nodes

    @cached_property
    def nodes_by_name(self) -> Dict[str, Node]:
        return {node.name: node for node in self._graph.nodes}

    @cached_property
    def links(self) -> OutLinkViewWrapper:
        """Returns a view on the links of the network."""
        return OutLinkViewWrapper(self._graph)

    @property
    def out_links(self) -> OutLinkViewWrapper:
        """Alias of the `links` property."""
        return self.links

    @cached_property
    def in_links(self) -> InLinkViewWrapper:
        """Returns a view on the inward links of the network."""
        return InLinkViewWrapper(self._graph)

    @cached_property
    def links_by_name(self) -> Dict[str, Link[VarType]]:
        return {  # type: ignore[var-annotated]
            link.name: link for _, _, link in self.links
        }

    @cached_property
    def nodes_by_link(self) -> Dict[Link[VarType], Tuple[Node, Node]]:
        return {  # type: ignore[var-annotated]
            link: (unode, dnode) for unode, dnode, link in self.links
        }

    @cached_property
    def origins(self) -> Dict[Origin[VarType], Node]:
        return {
            data[ORIGINENTRY]: node
            for node, data in self._graph.nodes.data()
            if ORIGINENTRY in data
        }

    @cached_property
    def origins_by_name(self) -> Dict[str, Origin[VarType]]:
        return {origin.name: origin for origin in self.origins}  # type: ignore[misc]

    @cached_property
    def origins_by_node(self) -> Dict[Node, Origin[VarType]]:
        d = self.origins
        return dict(zip(d.values(), d.keys()))  # type: ignore[arg-type]

    @cached_property
    def destinations(self) -> Dict[Destination[VarType], Node]:
        return {
            data[DESTINATIONENTRY]: node
            for node, data in self._graph.nodes.data()
            if DESTINATIONENTRY in data
        }

    @cached_property
    def destinations_by_name(self) -> Dict[str, Destination[VarType]]:
        return {
            destination.name: destination  # type: ignore[misc]
            for destination in self.destinations
        }

    @cached_property
    def destinations_by_node(self) -> Dict[Node, Destination]:
        d = self.destinations
        return dict(zip(d.values(), d.keys()))

    @property
    def elements(self) -> Iterable[ElementWithVars[VarType]]:
        """Gets an iterator to all the elements of the network."""
        return chain(
            (link[-1] for link in self.links),  # type: ignore[var-annotated]
            self.origins,  # type: ignore[arg-type]
            self.destinations,  # type: ignore[arg-type]
        )

    @property
    def states(self) -> Dict[ElementWithVars[VarType], Dict[str, VarType]]:
        """Gets the states of the network's elements."""
        return {
            el: el.states for el in self.elements if el.has_states  # type: ignore[misc]
        }

    @property
    def next_states(
        self,
    ) -> Dict[ElementWithVars[VarType], Dict[str, VarType]]:
        """Gets the states of the network's elements after stepping the
        dynamics for one time step."""
        return {
            el: el.next_states  # type: ignore[misc]
            for el in self.elements
            if el.has_next_states
        }

    @property
    def actions(self) -> Dict[ElementWithVars[VarType], Dict[str, VarType]]:
        """Gets the control action of the network's elements."""
        return {
            el: el.actions  # type: ignore[misc]
            for el in self.elements
            if el.has_actions
        }

    @property
    def disturbances(
        self,
    ) -> Dict[ElementWithVars[VarType], Dict[str, VarType]]:
        """Gets the disturbances of the network's elements."""
        return {
            el: el.disturbances  # type: ignore[misc]
            for el in self.elements
            if el.has_disturbances
        }

    @invalidate_cache(nodes_by_name)
    def add_node(self, node: Node) -> "Network":
        """Adds a node to the highway network.

        Parameters
        ----------
        node : Node
            Node to be added.

        Returns
        -------
        Network
            A reference to itself.
        """
        self._graph.add_node(node)
        return self

    @invalidate_cache(nodes_by_name)
    def add_nodes(self, nodes: Iterable[Node]) -> "Network":
        """Adds multiple nodes. See `Network.add_node`.

        Parameters
        ----------
        nodes : iterable of Nodes
            Nodes to be added.

        Returns
        -------
        Network
            A reference to itself.
        """
        self._graph.add_nodes_from(nodes)
        return self

    @invalidate_cache(links_by_name, nodes_by_link)
    def add_link(
        self, node_up: Node, link: Link[VarType], node_down: Node
    ) -> "Network":
        """Adds a link to the highway network, between two nodes.

        Parameters
        ----------
        node_up : Node
            Upstream node, that is, where traffic is coming from.
        link : Link
            The link to be added connecting the two nodes.
        node_down : Node
            Downstream node, that is, where traffic is going to.

        Returns
        -------
        Network
            A reference to itself.
        """
        self._graph.add_edge(node_up, node_down, **{LINKENTRY: link})
        return self

    @invalidate_cache(links_by_name, nodes_by_link)
    def add_links(self, links: Iterable[Tuple[Node, Link[VarType], Node]]) -> "Network":
        """Adds multiple links. See `Network.add_link`.

        Parameters
        ----------
        nodes : iterable of Tuple[Node, Link, Node]
            Links to be added between the corresponding nodes.

        Returns
        -------
        Network
            A reference to itself.
        """

        def get_edge(linkdata: Tuple[Node, Link[VarType], Node]):
            node_up, link, node_down = linkdata
            return (node_up, node_down, {LINKENTRY: link})

        self._graph.add_edges_from(get_edge(link) for link in links)
        return self

    @invalidate_cache(origins, origins_by_node, origins_by_name)
    def add_origin(self, origin: Origin[VarType], node: Node) -> "Network":
        """Adds the given traffic origin to the node.

        Parameters
        ----------
        origin : Origin
            Origin to be added to the network.
        node : Node
            Node which the origin is attached to.

        Returns
        -------
        Network
            A reference to itself.
        """
        if node not in self.nodes:
            self._graph.add_node(node, **{ORIGINENTRY: origin})
        else:
            self.nodes[node][ORIGINENTRY] = origin
        return self

    @invalidate_cache(destinations, destinations_by_node, destinations_by_name)
    def add_destination(
        self, destination: Destination[VarType], node: Node
    ) -> "Network":
        """Adds the given traffic destination to the node.

        Parameters
        ----------
        destination : Destination
            Destination to be added to the network.
        node : Node
            Node which the destination is attached to.

        Returns
        -------
        Network
            A reference to itself.
        """
        if node not in self.nodes:
            self._graph.add_node(node, **{DESTINATIONENTRY: destination})
        else:
            self.nodes[node][DESTINATIONENTRY] = destination
        self.destinations_by_name[
            destination.name
        ] = destination  # type: ignore[assignment]
        return self

    def add_path(
        self,
        path: Iterable[Union[Node, Link[VarType]]],
        origin: Optional[Origin[VarType]] = None,
        destination: Optional[Destination[VarType]] = None,
    ) -> "Network":
        """Adds a path of nodes and links between the origin and the
        destination.

        Parameters
        ----------
        path : iterable of either Nodes and Links
            A path consists of an alternating sequence of nodes and links,
            starting from the first node and ending at the last. For example, a
            valid path is: `node1, link1, node2, link2, node3, ..., nodeN`.
        origin : Origin, optional
            The origin where the path starts from. Pass `None` to have no
            origin attached to the first node in `path`.
        destination : Destination, optional
            The destination where the path ends in. Pass `None` to have no
            destination attached to the last node in `path`.

        Returns
        -------
        Network
            A reference to itself.

        Raises
        ------
        TypeError
            Raises if
            - the first or last points in `path` are not a `Node`
            - the alternation of `Link`s and `Node`s is not respected
            - the path has length 1, which is not accepted.
        """
        path = iter(path)
        first_node = next(path)
        if not isinstance(first_node, Node):
            raise TypeError(
                f"First element of the path must be a `{Node.__name__}`; got "
                f"{type(first_node)} instead."
            )
        self.add_node(first_node)
        if origin is not None:
            self.add_origin(origin, first_node)
        current_link: List[Union[Node, Link[VarType]]] = [first_node]

        longer_than_one = False
        for i, point in enumerate(path):
            longer_than_one = True
            current_link.append(point)
            L = len(current_link)
            if L == 2:
                if not isinstance(point, Link):
                    raise TypeError(
                        f"Expected a `{Link.__name__}` at index {i} of the path; "
                        f"got {type(point)} instead."
                    )
            else:  # L == 3
                if not isinstance(point, Node):
                    raise TypeError(
                        f"Expected a `{Node.__name__}` at index {i} of the path; "
                        f"got {type(point)} instead."
                    )
                self.add_node(point)
                self.add_link(*current_link)
                current_link = current_link[-1:]
        if not longer_than_one:
            raise ValueError("Path must be longer than a single node.")

        last_node = point
        if not isinstance(first_node, Node):
            raise TypeError(
                f"Last element of the path must be a `{Node.__name__}`; got "
                f"{type(first_node)} instead."
            )
        if destination is not None:
            self.add_destination(destination, last_node)
        return self

    def is_valid(self, raises: bool = False) -> Tuple[bool, List[str]]:
        """Checks whether the network is consistent.

        Parameters
        ----------
        raises : bool, optional
            If `True`, if an issue is found, an exception is raised. If
            `False`, then the exception messages are returned in a list. By
            default, `False`.

        Returns
        -------
        bool
            `True` if the network is valid, `False` otherwise.
        list[str]
            A list of messages describing the invalid issues found. Returned
            only if `raises=False`.

        Raises
        ------
        InvalidNetworkError
            Raises if `raises=True` and if
             1) a link, origin or destination is duplicated in the network

             2) a node has both an origin and a destination

             3) a node has no entering and no exiting links

             4) a node with no entering links has no origin

             5) a node with no exiting links has no destination

             6) a node with an origin (not a ramp) has also entering links

             7) a node with an origin has multiple exiting links

             8) a node with a destination has multiple entering links

             9) a node with a destination has also exiting links.
        """
        msgs = []

        def origin_destination_yielder():
            for data, entry in product(
                self._graph.nodes.values(), (ORIGINENTRY, DESTINATIONENTRY)
            ):
                if entry in data:
                    yield data[entry]

        # (1)
        count: Dict[object, int] = {}
        for o in chain(
            (link[2] for link in self.links), iter(origin_destination_yielder())
        ):
            d = count.get(o, 0) + 1
            if d > 1:
                msgs.append(f"Element {o.name} is duplicated in the network.")
                if raises:
                    raise InvalidNetworkError(msgs[-1])
            count[o] = d

        # (2); (3); (4); (5)
        for node, nodedata in self.nodes.data():
            if ORIGINENTRY in nodedata and DESTINATIONENTRY in nodedata:
                msgs.append(
                    f"Node {node.name} must either have an origin or a "
                    "destination, but not both."
                )
                if raises:
                    raise InvalidNetworkError(msgs[-1])
            n_in, n_out = len(self.in_links(node)), len(self.out_links(node))
            if n_in == 0 and n_out == 0:
                msgs.append(f"Node {node.name} is connected to no link.")
                if raises:
                    raise InvalidNetworkError(msgs[-1])
            if n_in == 0 and ORIGINENTRY not in nodedata:
                msgs.append(
                    f"Node {node.name} has neither any entering links nor an " "origin."
                )
                if raises:
                    raise InvalidNetworkError(msgs[-1])
            if n_out == 0 and DESTINATIONENTRY not in nodedata:
                msgs.append(
                    f"Node {node.name} has neither any exiting links nor a "
                    "destination."
                )
                if raises:
                    raise InvalidNetworkError(msgs[-1])

        # (6); (7)
        for origin, node in self.origins.items():
            if not isinstance(origin, MeteredOnRamp) and any(self.in_links(node)):
                msgs.append(
                    f"Expected node {node.name} to have no entering links, as "
                    f"it is connected to origin {origin.name} (only ramps "
                    "support entering links)."
                )
                if raises:
                    raise InvalidNetworkError(msgs[-1])
            if len(self.out_links(node)) > 1:
                msgs.append(
                    f"Expected node {node.name} to have at most one exiting "
                    f"link, as it is connected to origin {origin.name}."
                )
                if raises:
                    raise InvalidNetworkError(msgs[-1])

        # (8); (9)
        for destination, node in self.destinations.items():
            if len(self.in_links(node)) > 1:
                msgs.append(
                    f"Expected node {node.name} to have at most one entering "
                    "link, as it is connected to destination " + destination.name + "."
                )
                if raises:
                    raise InvalidNetworkError(msgs[-1])
            if any(self.out_links(node)):
                msgs.append(
                    f"Expected node {node.name} to have no exiting links, as "
                    f"it is connected to destination {destination.name}."
                )
                if raises:
                    raise InvalidNetworkError(msgs[-1])
        return not msgs, msgs

    def step(
        self,
        init_conditions: Optional[
            Dict[ElementWithVars[VarType], Dict[str, VarType]]
        ] = None,
        engine: Optional[EngineBase] = None,
        **other_parameters: VarType,
    ) -> None:
        """Steps the dynamics of the network's elements.

        Parameters
        ----------
        init_conditions : dict[element, dict[varname, variable]], optional
            For each element in the network, provides name-variable tuples to
            initialize states, actions and disturbances with specific values.
            These values must be compatible with the symbolic engine in type
            and shape. If not provided, variables are initialized
            automatically.
        engine : EngineBase, optional
            The engine to be used for stepping the dynamics. If `None`, the
            current engine is used.
        other_parameters : variables
            All the other parameters (e.g., sampling time) required during
            the computations.
        """
        # initialization
        if init_conditions is None:
            init_conditions = {}
        for el in self.elements:
            el.init_vars(
                init_conditions=init_conditions.get(el),  # type: ignore[arg-type]
                engine=engine,
            )

        # dynamics
        for origin in self.origins:
            origin.step(net=self, engine=engine, **other_parameters)
        for _, _, link in self.links:  # type: ignore[var-annotated]
            link.step(net=self, engine=engine, **other_parameters)



from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    Dict,
    Generator,
    Iterable,
    Optional,
    Tuple,
    Union,
)

import networkx as nx

from sym_metanet.util.types import VarType

if TYPE_CHECKING:
    from sym_metanet.blocks.links import Link
    from sym_metanet.blocks.nodes import Node


LINKENTRY = "link"
ORIGINENTRY = "origin"
DESTINATIONENTRY = "destination"


class OutLinkViewWrapper(nx.classes.reportviews.OutEdgeView):
    """Wrapper around `networkx`'s outward edge view to facilitate operations
    with link."""

    def __getitem__(self, e: Tuple["Node", "Node"]) -> "Link[VarType]":
        return super().__getitem__(e)[LINKENTRY]

    def __iter__(self) -> Generator[Tuple["Node", "Node", "Link[VarType]"], None, None]:
        for un, dns in self._nodes_nbrs():
            for dn, l in dns.items():
                yield (un, dn, l[LINKENTRY])

    def __call__(
        self,
        nbunch: Union[None, "Node", Iterable["Node"]] = None,
        data: Union[bool, str] = LINKENTRY,
        default: Optional[Dict[str, Any]] = None,
    ) -> Collection[Tuple["Node", "Node", "Link[VarType]"]]:
        return super().__call__(nbunch, data, default)


class InLinkViewWrapper(nx.classes.reportviews.InEdgeView):
    """Wrapper around `networkx`'s inward edge view to facilitate operations
    with link."""

    def __getitem__(self, e: Tuple["Node", "Node"]) -> "Link[VarType]":
        return super().__getitem__(e)[LINKENTRY]

    def __iter__(self) -> Generator[Tuple["Node", "Node", "Link[VarType]"], None, None]:
        for un, dns in self._nodes_nbrs():
            for dn, l in dns.items():
                yield (un, dn, l[LINKENTRY])

    def __call__(
        self,
        nbunch: Union[None, "Node", Iterable["Node"]] = None,
        data: Union[bool, str] = LINKENTRY,
        default: Optional[Dict[str, Any]] = None,
    ) -> Collection[Tuple["Node", "Node", "Link[VarType]"]]:
        return super().__call__(nbunch, data, default)



from abc import ABC, abstractmethod
from itertools import count
from typing import ClassVar, Dict, Generic, Optional, Set

from sym_metanet.util.types import VarType


class ElementBase:
    """Base class for any element for a highway modelled in METANET."""

    __slots__ = "name"
    __ids: Dict[type, count] = {}

    def __init__(self, name: Optional[str] = None) -> None:
        """Instantiates the element with the given `name` attribute.

        Parameters
        ----------
        name : str, optional
            Name of the element. If `None`, one is automatically created from a counter
            of the class' instancies.
        """
        cls = self.__class__
        if cls in self.__ids:
            _id = self.__ids[cls]
        else:
            _id = count(0)
            self.__ids[cls] = _id
        self.name = name or f"{cls.__name__}{next(_id)}"

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<{self.name}: {self.__class__.__name__}>"


class ElementWithVars(ElementBase, Generic[VarType], ABC):
    """Base class for any element with states, actions or disturbances."""

    __slots__ = ("states", "next_states", "actions", "disturbances")
    _states: ClassVar[Set[str]] = set()
    _actions: ClassVar[Set[str]] = set()
    _disturbances: ClassVar[Set[str]] = set()

    def __init__(self, name: Optional[str] = None) -> None:
        """Instantiates the element with the given `name` attribute.

        Parameters
        ----------
        name : str, optional
            Name of the element. If `None`, one is automatically created from a counter
            of the class' instancies.
        """
        super().__init__(name)
        self.states: Optional[Dict[str, VarType]] = None
        self.next_states: Optional[Dict[str, VarType]] = None
        self.actions: Optional[Dict[str, VarType]] = None
        self.disturbances: Optional[Dict[str, VarType]] = None

    @property
    def has_states(self) -> bool:
        """Gets whether this element has state variables."""
        return self.states is not None

    @property
    def has_next_states(self) -> bool:
        """Gets whether this element has state variables (computed by stepping the
        dynamics)."""
        return self.next_states is not None

    @property
    def has_actions(self) -> bool:
        """Gets whether this element has control action variables."""
        return self.actions is not None

    @property
    def has_disturbances(self) -> bool:
        """Gets whether this element has disturbance variables."""
        return self.disturbances is not None

    @abstractmethod
    def init_vars(self, *args, **kwargs) -> None:
        """Initializes the variable dicts (`states`, `actions`, `disturbances`) of this
        element."""
        raise NotImplementedError(
            f"Variable initialization not supported for {self.__class__.__name__}."
        )

    @abstractmethod
    def step_dynamics(self, *args, **kwargs) -> Dict[str, VarType]:
        """Internal method for stepping the element's dynamics by one time step.

        Returns
        -------
        Dict[str, VarType]
            A dict with the states at the next time step.

        Raises
        ------
        RuntimeError
            Raises if the shapes of the old and new states do not match.
        """
        raise NotImplementedError(
            f"Stepping the dynamics not supported for {self.__class__.__name__}."
        )

    def step(self, *args, **kwargs) -> None:
        """Steps the dynamics of this element."""
        assert self.states is not None, "States not initialized."
        next_states = self.step_dynamics(*args, **kwargs)
        if self.next_states is None:
            self.next_states = {}
        for name, state in self.states.items():
            next_state = next_states[name]
            self.next_states[name] = next_state
            if (
                hasattr(next_state, "shape")
                and hasattr(state, "shape")
                and next_state.shape != state.shape
            ):
                raise RuntimeError("Shapes of new and old states do not match.")

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<{self.name}: {self.__class__.__name__}>"



import unittest

import numpy as np

from sym_metanet import (
    CongestedDestination,
    Destination,
    Link,
    MeteredOnRamp,
    Origin,
    SimplifiedMeteredOnRamp,
    engines,
)

engine = engines.use("numpy", var_type="randn")


class TestLinks(unittest.TestCase):
    def test_init_vars__without_inital_condition__creates_vars(self):
        nb_seg = 4
        L = Link[np.ndarray](nb_seg, 3, 1, 180, 30, 100, 1.8)
        L.init_vars()
        self.assertIsNot(L.states, None)
        self.assertIs(L.actions, None)
        self.assertIs(L.disturbances, None)
        for n in ["rho", "v"]:
            self.assertIn(n, L.states)
            self.assertEqual(L.states[n].shape, (nb_seg,))

    def test_init_vars__with_inital_condition__copies_vars(self):
        nb_seg = 4
        L = Link[np.ndarray](nb_seg, 3, 1, 180, 30, 100, 1.8)
        init_conds = {"rho": np.random.rand(nb_seg), "v": np.random.rand(nb_seg)}
        L.init_vars(init_conds)
        self.assertIsNot(L.states, None)
        self.assertIs(L.actions, None)
        self.assertIs(L.disturbances, None)
        for n in ["rho", "v"]:
            self.assertIn(n, L.states)
            self.assertEqual(L.states[n].shape, (nb_seg,))
            np.testing.assert_equal(init_conds[n], L.states[n])


class TestDestinations(unittest.TestCase):
    def test_init_vars__no_value_is_initialized(self):
        D = Destination()
        D.init_vars()
        self.assertIs(D.states, None)
        self.assertIs(D.actions, None)
        self.assertIs(D.disturbances, None)


class TestCongestedDestinations(unittest.TestCase):
    def test_init_vars__without_inital_condition__creates_vars(self):
        # sourcery skip: class-extract-method
        D = CongestedDestination[np.ndarray]()
        D.init_vars()
        self.assertIs(D.states, None)
        self.assertIs(D.actions, None)
        self.assertIsNot(D.disturbances, None)
        self.assertIn("d", D.disturbances)
        self.assertIn(D.disturbances["d"].shape, {(1,), ()})

    def test_init_vars__with_inital_condition__copies_vars(self):
        D = CongestedDestination[np.ndarray]()
        init_conds = {"d": np.random.rand(1)}
        D.init_vars(init_conds)
        self.assertIs(D.states, None)
        self.assertIs(D.actions, None)
        self.assertIsNot(D.disturbances, None)
        self.assertIn("d", D.disturbances)
        self.assertIn(D.disturbances["d"].shape, {(1,), ()})
        np.testing.assert_equal(init_conds["d"], D.disturbances["d"])


class TestOrigins(unittest.TestCase):
    def test_init_vars__no_value_is_initialized(self):
        origin = Origin()
        origin.init_vars()
        self.assertIs(origin.states, None)
        self.assertIs(origin.actions, None)
        self.assertIs(origin.disturbances, None)


class TestMeteredOnRamp(unittest.TestCase):
    def test_init_vars__without_inital_condition__creates_vars(self):
        R = MeteredOnRamp[np.ndarray](1e5)
        R.init_vars()
        self.assertIsNot(R.states, None)
        self.assertIsNot(R.actions, None)
        self.assertIsNot(R.disturbances, None)
        for n, d in [("w", R.states), ("r", R.actions), ("d", R.disturbances)]:
            self.assertIn(n, d)
            self.assertIn(d[n].shape, {(1,), ()})

    def test_init_vars__with_inital_condition__copies_vars(self):
        R = MeteredOnRamp[np.ndarray](1e5)
        init_conds = {
            "w": np.random.rand(1),
            "r": np.random.rand(1),
            "d": np.random.rand(1),
        }
        R.init_vars(init_conds)
        self.assertIsNot(R.states, None)
        self.assertIsNot(R.actions, None)
        self.assertIsNot(R.disturbances, None)
        for n, d in [("w", R.states), ("r", R.actions), ("d", R.disturbances)]:
            self.assertIn(n, d)
            self.assertIn(d[n].shape, {(1,), ()})
            np.testing.assert_equal(init_conds[n], d[n])


class TestSimpleMeteredOnRamp(unittest.TestCase):
    def test_init_vars__without_inital_condition__creates_vars(self):
        R = SimplifiedMeteredOnRamp[np.ndarray](1e5)
        R.init_vars()
        self.assertIsNot(R.states, None)
        self.assertIsNot(R.actions, None)
        self.assertIsNot(R.disturbances, None)
        for n, d in [("w", R.states), ("q", R.actions), ("d", R.disturbances)]:
            self.assertIn(n, d)
            self.assertIn(d[n].shape, {(1,), ()})

    def test_init_vars__with_inital_condition__copies_vars(self):
        R = SimplifiedMeteredOnRamp[np.ndarray](1e5)
        init_conds = {
            "w": np.random.rand(1),
            "q": np.random.rand(1),
            "d": np.random.rand(1),
        }
        R.init_vars(init_conds)
        self.assertIsNot(R.states, None)
        self.assertIsNot(R.actions, None)
        self.assertIsNot(R.disturbances, None)
        for n, d in [("w", R.states), ("q", R.actions), ("d", R.disturbances)]:
            self.assertIn(n, d)
            self.assertIn(d[n].shape, {(1,), ()})
            np.testing.assert_equal(init_conds[n], d[n])


if __name__ == "__main__":
    unittest.main()



from typing import TYPE_CHECKING, Collection, Dict, Optional, Tuple

from sym_metanet.blocks.base import ElementWithVars
from sym_metanet.engines.core import EngineBase, get_current_engine
from sym_metanet.util.funcs import first
from sym_metanet.util.types import VarType

if TYPE_CHECKING:
    from sym_metanet.blocks.links import Link
    from sym_metanet.blocks.nodes import Node
    from sym_metanet.network import Network


class Destination(ElementWithVars[VarType]):
    """Ideal congestion-free destination, representing a sink where cars can leave the
    highway with no congestion (i.e., no slowing down due to downstream density)."""

    def init_vars(self, *args, **kwargs) -> None:
        """Initializes no variable in the ideal destination."""

    def step_dynamics(self, *args, **kwargs) -> Dict[str, VarType]:
        """No dynamics to steps in the ideal destination."""
        return {}

    def get_density(self, net: "Network", **kwargs) -> VarType:
        """Computes the (downstream) density induced by the ideal destination.

        Parameters
        ----------
        net : Network
            The network this destination belongs to.

        Returns
        -------
        symbolic variable
            The destination's downstream density.
        """
        return self._get_entering_link(net).states["rho"][-1]

    def _get_entering_link(self, net: "Network") -> "Link[VarType]":
        """Internal utility to fetch the link entering this destination (can only be
        one)."""
        links_up: Collection[Tuple["Node", "Node", "Link[VarType]"]] = net.in_links(
            net.destinations[self]  # type: ignore[index]
        )
        assert (
            len(links_up) == 1
        ), "Internal error. Only one link can enter a destination."
        return first(links_up)[-1]


class CongestedDestination(Destination[VarType]):
    """Destination with a downstream density scenario to emulate congestions, that is,
    cars cannot exit freely the highway but must slow down and, possibly, create a
    congestion."""

    _disturbances = {"d"}

    def init_vars(
        self,
        init_conditions: Optional[Dict[str, VarType]] = None,
        engine: Optional[EngineBase] = None,
    ) -> None:
        """Initializes
        - `d`: downstream density scenario (disturbance).

        Parameters
        ----------
        net : Network
            The network this destination belongs to.
        init_conditions : dict[str, variable], optional
            Provides name-variable tuples to initialize states, actions and disturbances
            with specific values. These values must be compatible with the symbolic
            engine in type and shape. If not provided, variables are initialized
            automatically.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.
        """
        if engine is None:
            engine = get_current_engine()
        self.disturbances: Dict[str, VarType] = {
            "d": engine.var(f"d_{self.name}")
            if init_conditions is None or "d" not in init_conditions
            else init_conditions["d"]
        }

    def get_density(
        self, net: "Network", engine: Optional[EngineBase] = None, **kwargs
    ) -> VarType:
        """Computes the (downstream) density induced by the congested destination.

        Parameters
        ----------
        net : Network
            The network this destination belongs to.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        variable
            The destination's downstream density.
        """
        if engine is None:
            engine = get_current_engine()
        link_up = self._get_entering_link(net)
        return engine.destinations.get_congested_downstream_density(
            link_up.states["rho"][-1], self.disturbances["d"], link_up.rho_crit
        )



from typing import TYPE_CHECKING, Collection, Optional, Tuple

from sym_metanet.blocks.base import ElementBase
from sym_metanet.engines.core import EngineBase, get_current_engine
from sym_metanet.util.funcs import first
from sym_metanet.util.types import Variable, VarType

if TYPE_CHECKING:
    from sym_metanet.blocks.links import Link
    from sym_metanet.network import Network


class Node(ElementBase):
    """Node of the highway [1, Section 3.2.2] representing, e.g., the connection between
    two links. Nodes do not correspond to actual physical components of the highway, but
    are used to separate links in case there is a major change in the link parameters or
    a junction or bifurcation.

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    def get_downstream_density(
        self, net: "Network", engine: Optional[EngineBase] = None, **kwargs
    ) -> Variable:
        """Computes the (virtual) downstream density of the node.

        Parameters
        ----------
        net : Network
            The network which node and link belongs to.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        symbolic variable
            Returns the (virtual) downstream density.
        """
        if engine is None:
            engine = get_current_engine()

        # following the link entering this node, this node can only be a
        # destination or have multiple exiting links
        if self in net.destinations_by_node:
            return net.destinations_by_node[self].get_density(
                net, engine=engine, **kwargs
            )

        # if no destination, then there must be 1 or more exiting links
        links_down: Collection[Tuple["Node", "Node", "Link[Variable]"]] = net.out_links(
            self
        )
        if len(links_down) == 1:
            return first(links_down)[-1].states["rho"][0]
        rho_firsts = engine.vcat(
            *(dlink.states["rho"][-1] for _, _, dlink in links_down)
        )
        return engine.nodes.get_downstream_density(rho_firsts)

    def get_upstream_speed_and_flow(
        self,
        net: "Network",
        link: "Link[VarType]",
        engine: Optional[EngineBase] = None,
        **kwargs
    ) -> Tuple[VarType, VarType]:
        """Computes the (virtual) upstream speed and flow of the node for this the
        current link.

        Parameters
        ----------
        net : Network
            The network which node and link belongs to.
        link : Link
            The current link (which departs from this node) querying this information
            from the node.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        tuple[symbolic variable, symbolic variable]
            Returns the (virtual) upstream speed and flow.
        """
        if engine is None:
            engine = get_current_engine()

        # the node can have 1 or more entering links, as well as a ramp origin.
        # Speed is dictated by the entering links, if any; otherwise by the
        # origin (same as first segment). Flow is dictated both by entering
        # links and origin.
        links_up: Collection[Tuple["Node", "Node", "Link[VarType]"]] = net.in_links(
            self
        )
        n_up = len(links_up)
        if self in net.origins_by_node:
            origin = net.origins_by_node[self]
            v_o = origin.get_speed(net, engine=engine, **kwargs)
            q_o = origin.get_flow(net, engine=engine, **kwargs)
        else:
            v_o = None
            q_o = None

        if n_up == 0:
            v = v_o
            q = q_o
        elif n_up == 1:
            link_up = next(iter(links_up))[-1]
            v = link_up.states["v"][-1]
            q = link_up.get_flow(engine)[-1]
            if q_o is not None:
                q += q_o  # type: ignore[assignment,operator]
        else:
            v_last = []
            q_last = []
            for _, _, link_up in links_up:
                v_last.append(link_up.states["v"][-1])
                q_last.append(link_up.get_flow(engine)[-1])
            v_last = engine.vcat(*v_last)
            q_last = engine.vcat(*q_last)
            links_down: Collection[
                Tuple["Node", "Node", "Link[VarType]"]
            ] = net.out_links(self)
            betas = engine.vcat(*(dlink.turnrate for _, _, dlink in links_down))

            v = engine.nodes.get_upstream_speed(q_last, v_last)
            q = engine.nodes.get_upstream_flow(q_last, link.turnrate, betas, q_o)
        return v, q  # type: ignore[return-value]



__all__ = [
    "InvalidNetworkError",
    "EngineNotFoundWarning",
    "EngineNotFoundError",
    "engines",
    "Node",
    "Link",
    "Origin",
    "MeteredOnRamp",
    "SimplifiedMeteredOnRamp",
    "Destination",
    "CongestedDestination",
    "Network",
]

import sym_metanet.engines as engines
from sym_metanet.errors import (
    EngineNotFoundError,
    EngineNotFoundWarning,
    InvalidNetworkError,
)

_notfound = True
for _engine in engines.get_available_engines().keys():
    try:
        engine = engines.use(_engine)
        _notfound = False
        break
    except ImportError:
        continue
if _notfound:
    import warnings

    warnings.warn("No available symbolic engine found.", EngineNotFoundWarning)
del _notfound, _engine

from sym_metanet.blocks.destinations import CongestedDestination, Destination
from sym_metanet.blocks.links import Link
from sym_metanet.blocks.nodes import Node
from sym_metanet.blocks.origins import MeteredOnRamp, Origin, SimplifiedMeteredOnRamp
from sym_metanet.network import Network



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Dict, Literal, Type, Union

import sym_metanet
from sym_metanet.errors import EngineNotFoundError

if TYPE_CHECKING:
    from sym_metanet.network import Network


class NodesEngineBase(ABC):
    """Abstract class of a symbolic engine for modelling highway nodes via the METANET
    framework. The methods of this class implement the various equations proposed in the
    framework, which can be found in [1].

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    @staticmethod
    @abstractmethod
    def get_upstream_flow(q_lasts, beta, betas, q_orig=0):
        """Computes the upstream flow from a node to a given exiting link, where the
        node can have multiple entering links and multiple exiting links, according to
        [1, Section 3.2.2].

        Parameters
        ----------
        q_lasts
            Flows of the last segments of the links entering the node.
        beta
            Turnrate of the given link exiting from the node.
        betas
            Turnrates of all the exiting links (including this link).
        q_orig : optional
            An optional upstream flow from an origin.

        Returns
        -------
        flow
            (virtual) upstream flow in the first segment of the exiting link.
        """

    @staticmethod
    @abstractmethod
    def get_upstream_speed(q_lasts, v_lasts):
        """Computes the upstream speed from a node to a given exiting link, where the
        node can have multiple entering links and multiple exiting links, according to
        [1, Equation 3.10].

        Parameters
        ----------
        q_lasts
            Flows of the last segments of the links entering the node.
        v_lasts
            Speeds of the last segments of the links entering the node.

        Returns
        -------
        speed
            (virtual) upstream speed in the first segment of the exiting link.
        """

    @staticmethod
    @abstractmethod
    def get_downstream_density(rho_firsts):
        """Computes the downstream density from a node to a given entering link, where
        the node can have multiple entering links and multiple exiting links, according
        to [1, Equation 3.9].

        Parameters
        ----------
        rho_firsts
            Densities of the first segments of the links exiting the node.

        Returns
        -------
        speed
            (virtual) downstream density in the last segment of the entering link.
        """


class LinksEngineBase(ABC):
    """Abstract class of a symbolic engine for modelling highway link via the METANET
    framework. The methods of this class implement the various equations proposed in the
    framework, which can be found in [1].

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    @staticmethod
    @abstractmethod
    def get_flow(rho, v, lanes):
        """Computes the flows of the link's segments, according to [1, Equation 3.1].

        Parameters
        ----------
        rho
            Densities of the link's segments.
        v_free
            Speeds of the link's segments.
        lanes
            Number of lanes in the link.

        Returns
        -------
        flow
            The flow in each link's segment.
        """

    @staticmethod
    @abstractmethod
    def step_density(rho, q, q_up, lanes, L, T):
        """Computes the densities of the link's segments at the next time instant,
        according to [1, Equation 3.2].

        Parameters
        ----------
        rho
            Densities of this link's segments.
        q
            Flows of this link's segments.
        q_up
            Flows of the upstream link's segments.
        lanes
            Number of lanes in the link.
        L
            Link's segment length.
        T
            Sampling time.

        Returns
        -------
        densities_next
            The density in each link's segment at the next time instant.
        """

    @staticmethod
    @abstractmethod
    def step_speed(
        v,
        v_up,
        rho,
        rho_down,
        Veq,
        lanes,
        L,
        tau,
        eta,
        kappa,
        T,
        q_ramp=None,
        delta=None,
        lanes_drop=None,
        phi=None,
        rho_crit=None,
    ):
        """Computes the speeds of the link's segments at the next time instant,
        according to [1, Equation 3.3]. If `q_ramp` and `delta` are provided, then it
        also accounts for merging phenomum [1, Equation 3.7]. Similarly, if `lane_drop`
        and `phi` are provided, then it also accounts for lane drops in the incoming
        link [1, Equation 3.8].

        Parameters
        ----------
        v
            Speeds of this link's segments.
        v_up
            Speeds of the upstream link's segments.
        rho
            Densities of this link's segments.
        rho_down
            Densities of the downstream link's segments.
        Veq
            Equilibrium speed of this link.
        lanes
            Number of lanes in the link.
        L
            Link's segment length.
        tau
            Model parameter for the relaxation term.
        eta
            Model parameter for the anticipation term.
        kappa
            Model parameter for the anticipation term.
        T
            Sampling time.
        q_ramp : optional
            If provided, it represents the flow of an on-ramp attached to the first
            segment of this link, and is used to compute merging phenomenum of on-ramp
            traffic with link traffic.
        delta : optional
            Model parameter for merging phenomenum. Necessary only if `q_ramp` is
            provided as well.
        lanes_drop : optional
            If provided, it represents the difference of lanes of this link minus the
            lanes in the next link, and is used to compute the impact of the lane drop
            on speeds.
        phi : optional
            Model parameter for lane drop phenomenum. Necessary only if `lanes_drop` and
            `rho_crit` are provided as well.
        rho_crit : optional
            Critical density of the link. Necessary only if `lanes_drop` and `phi` are
            provided as well.

        Returns
        -------
        speeds_next
            The speed in each link's segment at the next time instant.
        """

    @staticmethod
    @abstractmethod
    def Veq(rho, v_free, rho_crit, a):
        """Computes the equilibrium speed of the link, according to [1, Equation 3.4].

        Parameters
        ----------
        rho
            Densities of the link's segments.
        v_free
            Free-flow speed of the link.
        rho_crit
            Critical density of the link.
        a
            Model parameter in the equilibrium speed exponent.

        Returns
        -------
        Veq
            The equilibrium speed of the link.
        """


class OriginsEngineBase(ABC):
    """Abstract class of a symbolic engine for modelling highway origins via the METANET
    framework. The methods of this class implement the various equations proposed in the
    framework, which can be found in [1].

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    @staticmethod
    @abstractmethod
    def step_queue(w, d, q, T):
        """Computes the queue of the origin at the next time instant, according to
        [1, Section 3.2.1].

        Parameters
        ----------
        w
            Queue of this origin.
        d
            Demand at this origin.
        q
            Outflow from this origin.
        T
            Sampling time.

        Returns
        -------
        queue_next
            The queue of the origin at the next time instant.
        """

    @staticmethod
    @abstractmethod
    def get_ramp_flow(
        d, w, C, r, rho_max, rho_first, rho_crit, T, type: Literal["in", "out"] = "out"
    ):
        """Computes the flows of the ramp origin according to [1, Equation 3.5] if
        `type='in'`, or [1, Equation 3.6] if `type='out'`.

        Parameters
        ----------
        d
            Demand at the ramp.
        w
            Queue of the ramp.
        C
            Capacity of the ramp.
        r
            Metering rate (control action) of the ramp.
        rho_max
            Maximum density of the link the ramp is attached to.
        rho_first
            Density of the first segment of the link the ramp is attached to.
        rho_crit
            Critical density of the link the ramp is attached to.
        T
            Sampling time.
        type : 'in' or 'out', optional
            Whether the metering rate `r` should be inside or outside the min function;
            by default, 'out'. See [1, Equations 3.5 and 3.6] for more details.

        Returns
        -------
        flow
            The flow of the ramp.
        """

    @staticmethod
    @abstractmethod
    def get_simplifiedramp_flow(
        qdes,
        d=None,
        w=None,
        C=None,
        rho_max=None,
        rho_first=None,
        rho_crit=None,
        T=None,
        type: Literal["limited", "unlimited"] = "limited",
    ):
        """Computes the flows of a simplified ramp origin.

        Parameters
        ----------
        qdes
            Desired flow at the ramp.
        d : optional
            Demand at the ramp.
        w : optional
            Queue of the ramp.
        C : optional
            Capacity of the ramp.
        rho_first : optional
            Density of the first segment of the link the ramp is attached to. To be
            provided only if `type=limited`.
        rho_max : optional
            Maximum density of the link the ramp is attached to. To be provided only if
            `type=limited`.
        rho_crit : optional
            Critical density of the link the ramp is attached to. To be provided only if
            `type=limited`.
        T : optional
            Sampling time.
        type : "limited" or "unlimited", optional
            Type of equation to compute the flow for a simplified metered ramp. By
            default, "limited" is selected, which means that the desired flow is limited
            according to the capacity, current demands, queues and densities of the ramp
            and its attached segment. This equation needs that all optional arguments
            are provided. Otherwise, "unlimited" makes sure that the  desired flow is
            the actual flow of the ramp, which can cause negative queues though.

        Returns
        -------
        flow
            The flow of the ramp.
        """


class DestinationsEngineBase(ABC):
    """Abstract class of a symbolic engine for modelling highway destinations via the
    METANET framework. The methods of this class implement the various equations
    proposed in the framework, which can be found in [1].

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    @staticmethod
    @abstractmethod
    def get_congested_downstream_density(rho_last, rho_destination, rho_crit):
        """For a link with a downstream congested destination, returns the downstream
        density.

        Parameters
        ----------
        rho_last
            Density in the link's last segment.
        rho_destination
            Density scenario of the congested destination.
        rho_crit
            Critical density of the link entering the destination.

        Returns
        -------
        density
            The downstream density for a link connected to a destination.
        """


class EngineBase(ABC):
    """Abstract class of a symbolic engine for modelling highways via the METANET
    framework. The methods of this class implement the various equations proposed in the
    framework, which can be found in [1].

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    @property
    @abstractmethod
    def nodes(self) -> Type[NodesEngineBase]:
        pass

    @property
    @abstractmethod
    def links(self) -> Type[LinksEngineBase]:
        pass

    @property
    @abstractmethod
    def origins(self) -> Type[OriginsEngineBase]:
        pass

    @property
    @abstractmethod
    def destinations(self) -> Type[DestinationsEngineBase]:
        pass

    @abstractmethod
    def var(self, name: str, n: int = 1, *args, **kwargs):
        """Creates a variable.

        Parameters
        ----------
        name : str
            Name of the variable.
        n : int
            Length of the variable.

        Returns
        -------
        sym variable
            The symbolic variable.
        """

    @abstractmethod
    def vcat(self, *arrays):
        """Concatenates vertically two elements/variables.

        Parameters
        ----------
        arrays
            Values to be vertically concatenated.

        Returns
        -------
        sym variable
            A unique variable resulting from the vertical concatenation.
        """

    @abstractmethod
    def to_function(self, net: "Network", *args, **kwargs) -> Callable:
        """Returns the network's dynamics as a callable function, with different nature
        depending on the engine implementation.

        Parameters
        ----------
        net : Network
            The network whose dynamics must be translated into a function.

        Returns
        -------
        Callable
            The function representing the dynamics.
        """


def get_current_engine() -> EngineBase:
    """Gets the current symbolic engine.

    Returns
    -------
    EngineBase
        The current symbolic engine.
    """
    return sym_metanet.engine


def get_available_engines() -> Dict[str, Dict[str, str]]:
    """Returns the available symbolic engines for METANET modelling, for which an
    implementation exists.

    Returns
    -------
    Dict[str, Dict[str, str]]
        The available engines in the form of a `dict` whose keys are the available
        engine class types, and the values are info on each engine, such as module and
        class name.
    """
    return {
        "casadi": {"module": "sym_metanet.engines.casadi", "class": "Engine"},
        "numpy": {"module": "sym_metanet.engines.numpy", "class": "Engine"},
    }


def use(engine: Union[str, EngineBase], *args, **kwargs) -> EngineBase:
    """Uses the given symbolic engine for computations.

    Parameters
    ----------
    engine : str or instance of engine
        If a string, then `engine` must represent the class name of one of the available
        engines, so that it can be instantiated with args and kwargs (see
        `sym_metanet.engines.get_available_engines`). Otherwise, it must be an instance
        of an engine itself that inherits from `sym_metanet.engines.EngineBase`.
    args, kwargs
        Passed to the engine constructor in case `engine` is a string.

    Returns
    -------
    EngineBase
        A reference to the new engine, if `engine` is a string, or the reference to the
        same instance, otherwise.

    Raises
    ------
    ValueError
        Raises in case `engine` is neither a string nor a `EngineBase` instance.
    EngineNotFoundError
        Raises in case `engine` is a string but matches no available engines.
    """
    if isinstance(engine, EngineBase):
        sym_metanet.engine = engine
    else:
        engines = get_available_engines()
        if engine not in engines:
            raise EngineNotFoundError(
                f'Engine class must be in {{{", ".join(engines)}}}; got '
                f"{engine} instead."
            )
        from importlib import import_module

        engineinfo = engines[engine]
        cls = getattr(import_module(engineinfo["module"]), engineinfo["class"])
        sym_metanet.engine = cls(*args, **kwargs)
    return sym_metanet.engine



import unittest

import casadi as cs
import numpy as np
from parameterized import parameterized

import sym_metanet as metanet
from sym_metanet import (
    CongestedDestination,
    Link,
    MeteredOnRamp,
    Network,
    Node,
    SimplifiedMeteredOnRamp,
    engines,
)
from sym_metanet.engines.casadi import Engine as CasadiEngine
from sym_metanet.engines.numpy import Engine as NumpyEngine
from sym_metanet.errors import EngineNotFoundError
from sym_metanet.util.funcs import first


def get_net(link_with_ramp: int = 2):
    assert link_with_ramp in {1, 2}
    L = 1
    lanes = 2
    C = (3500, 2000)
    tau = 18 / 3600
    kappa = 40
    eta = 60
    rho_max = 180
    delta = 0.0122
    T = 10 / 3600
    a_sym = cs.SX.sym("a")
    v_free_sym = cs.SX.sym("v_free")
    rho_crit_sym = cs.SX.sym("rho_crit_sym")
    N1 = Node(name="N1")
    N2 = Node(name="N2")
    N3 = Node(name="N3")
    N4 = Node(name="N3")
    L1 = Link(1, lanes, L, rho_max, rho_crit_sym, v_free_sym, a_sym, name="L1")
    L2 = Link(1, lanes, L, rho_max, rho_crit_sym, v_free_sym, a_sym, name="L2")
    L3 = Link(1, lanes, L, rho_max, rho_crit_sym, v_free_sym, a_sym, name="L3")
    O1 = MeteredOnRamp(C[0], name="O1")
    O2 = SimplifiedMeteredOnRamp(C[1], name="O2")
    D1 = CongestedDestination(name="D1")
    net = (
        Network(name="A1")
        .add_path(origin=O1, path=(N1, L1, N2, L2, N3, L3, N4), destination=D1)
        .add_origin(O2, N2 if link_with_ramp == 1 else N3)
    )
    sym_pars = {"rho_crit": rho_crit_sym, "a": a_sym, "v_free": v_free_sym}
    others = {
        "L": L,
        "lanes": lanes,
        "C": C,
        "rho_max": rho_max,
        "T": T,
        "tau": tau,
        "eta": eta,
        "kappa": kappa,
        "delta": delta,
    }
    return net, sym_pars, others


def get_hardcoded_dynamics(
    L, lanes, C, tau, kappa, eta, rho_max, delta, T, link_with_ramp: int = 2, **kwargs
) -> cs.Function:
    # sourcery skip: use-contextlib-suppress
    assert link_with_ramp in {1, 2}
    rho = cs.SX.sym("rho", 3, 1)
    v = cs.SX.sym("v", 3, 1)
    w = cs.SX.sym("w", 2, 1)
    r = cs.SX.sym("r")
    q_O2 = cs.SX.sym("q_O2")
    d = cs.SX.sym("d", 3, 1)
    rho_crit = cs.SX.sym("rho_crit")
    a = cs.SX.sym("a")
    v_free = cs.SX.sym("v_free")

    q_O1 = r * cs.fmin(
        d[0] + w[0] / T, C[0] * cs.fmin(1, (rho_max - rho[0]) / (rho_max - rho_crit))
    )
    q_O2_limited = cs.mmin(
        cs.vertcat(  # must limit the desired q_O2 to be feasible
            q_O2,
            w[1] / T + d[1],
            C[1],
            C[1] * (rho_max - rho[1]) / (rho_max - rho_crit),
        ),
    )
    q_o = cs.vertcat(q_O1, q_O2_limited)
    w_next = w + T * (d[:2] - q_o)

    q = lanes * rho * v
    q_up = cs.vertcat(q_O1, q[0], q[1])
    q_up[link_with_ramp] += q_O2_limited
    v_up = cs.vertcat(v[0], v[0], v[1])
    rho_down = cs.vertcat(rho[1], rho[2], cs.fmax(cs.fmin(rho[2], rho_crit), d[2]))
    Veq = v_free * cs.exp((-1 / a) * ((rho / rho_crit)) ** a)
    rho_next = rho + (T / (L * lanes)) * (q_up - q)
    v_next = (
        v
        + T / tau * (Veq - v)
        + T / L * v * (v_up - v)
        - eta * T / tau / L * (rho_down - rho) / (rho + kappa)
    )
    merging_num = delta * T * q_O2_limited * v[link_with_ramp]
    merging_den = L * lanes * (rho[link_with_ramp] + kappa)
    v_next[link_with_ramp] -= merging_num / merging_den

    pars = cs.vertcat(rho_crit, a, v_free)
    # rho_next = cs.fmax(0, rho_next)
    v_next = cs.fmax(0, v_next)
    # w_next = cs.fmax(0, w_next)
    # q = cs.fmax(0, q)
    # q_o = cs.fmax(0, q_o)
    return cs.Function(
        "F", [rho, v, w, r, q_O2, d, pars], [rho_next, v_next, w_next, q, q_o]
    )


class TestEngines(unittest.TestCase):
    def test_get_current_engine__gets_correct_engine(self):
        self.assertIs(engines.get_current_engine(), metanet.engine)

    def test_get_available_engines__gets_available_engines_correctly(self):
        engines_ = engines.get_available_engines()
        self.assertIsInstance(engines_, dict)
        self.assertIsInstance(first(engines_.keys()), str)

    def test_use__sets_the_engine_correctly(self):
        old_engine = metanet.engine
        new_engine = engines.use("casadi")
        self.assertIsNot(metanet.engine, old_engine)
        self.assertIs(metanet.engine, new_engine)

    def test_use__raises__when_engine_not_found(self):
        invalid_engine = object()
        with self.assertRaises(EngineNotFoundError):
            engines.use(invalid_engine)


class TestCasadiEngine(unittest.TestCase):
    def test_to_function__fails__with_uninit_vars(self):
        net, sym_pars, other_pars = get_net()
        self.assertTrue(net.is_valid(raises=False)[0])
        engine = engines.use("casadi", sym_type="SX")

        with self.assertRaises(RuntimeError):
            engine.to_function(
                net=net,
                parameters=sym_pars,
                T=other_pars["T"],
                compact=1,
                more_out=True,
            )

    def test_to_function__fails__with_unstepped_dynamics(self):
        net, sym_pars, other_pars = get_net()
        self.assertTrue(net.is_valid(raises=False)[0])
        engine = engines.use("casadi", sym_type="SX")
        for el in net.elements:
            el.init_vars(engine=engine)
        with self.assertRaises(RuntimeError):
            engine.to_function(
                net=net,
                parameters=sym_pars,
                T=other_pars["T"],
                compact=1,
                more_out=True,
            )

    @parameterized.expand([(1,), (2,)])
    def test_to_function__numerically_works(self, link_with_ramp: int):
        net, sym_pars, other_pars = get_net(link_with_ramp)
        net.is_valid(raises=True)
        net.step(**other_pars)
        Fact: cs.Function = metanet.engine.to_function(
            net=net,
            **other_pars,
            parameters=sym_pars,
            more_out=True,
            force_positive_speed=True,
            compact=1,
        )
        Fexp = get_hardcoded_dynamics(**other_pars, link_with_ramp=link_with_ramp)

        np_random = np.random.Generator(np.random.PCG64(69420))
        names = Fact.name_out()
        for i in range(1_000):
            if i == 0:
                rho = np.asarray([10, 10, 10])
                v = np.asarray([100, 100, 100])
                w = np.asarray([0, 0])
                r = 1.0
                q = 575.0
                d = np.asarray([1027.85313984, 391.81732715, 21.76034038])
                a = 1.867
                v_free = 102.0
                rho_crit = 33.5
            else:
                rho = np_random.normal(30, 10, 3)
                v = np_random.normal(100, 30, 3)
                w = np_random.normal(100, 50, 2)
                r = np_random.random()
                q = np_random.normal(800, 200)
                d = np_random.normal((1e3, 1e3, 50), (200, 200, 20))
                rho_crit = np_random.uniform(10, 30)
                a = np_random.uniform(1, 2)
                v_free = np_random.uniform(40, 80)
            args = [
                np.maximum(0, arg)
                for arg in [rho, v, w, r, q, d, [rho_crit, a, v_free]]
            ]
            out_exp = Fexp(*args)
            out_act = Fact(*args)

            for name, exp_, act_ in zip(names, out_exp, out_act):
                exp: np.ndarray = exp_.full().flatten()
                act: np.ndarray = act_.full().flatten()
                self.assertTrue((exp >= 0).all(), f"[exp] {name} negative (iter {i})")
                self.assertTrue((act >= 0).all(), f"[act] {name} negative (iter {i})")
                np.testing.assert_allclose(
                    act,
                    exp,
                    atol=1e-6,
                    rtol=1e-6,
                    err_msg=f"{name} dissimilar (iter {i})",
                )


class TestCasadiVsNumpyEngine(unittest.TestCase):
    N = 7
    NE = NumpyEngine()
    CE = CasadiEngine[cs.DM]()

    def test_nodes__get_upstream_flow(self):
        N = self.N
        q_lasts = np.maximum(0, np.random.randn(N) * 50 + 200)
        betas = np.random.randint(low=1, high=10, size=N)
        q_orig = np.maximum(0, np.random.randn() * 50 + 200)
        args = [q_lasts, betas[0], betas, q_orig]
        np.testing.assert_allclose(
            self.NE.nodes.get_upstream_flow(*args),
            self.CE.nodes.get_upstream_flow(*map(cs.DM, args)).full().squeeze(),
        )

    def test_nodes__get_upstream_speed(self):
        N = self.N
        q_lasts = np.maximum(0, np.random.randn(N) * 50 + 200)
        v_lasts = np.maximum(0, np.random.randn(N) * 10 + 60)
        args = [q_lasts, v_lasts]
        np.testing.assert_allclose(
            self.NE.nodes.get_upstream_speed(*args),
            self.CE.nodes.get_upstream_speed(*map(cs.DM, args)).full().squeeze(),
        )

    def test_nodes__get_downstream_density(self):
        N = self.N
        rho_firsts = np.maximum(0, np.random.randn(N) * 5 + 30)
        np.testing.assert_allclose(
            self.NE.nodes.get_downstream_density(rho_firsts),
            self.CE.nodes.get_downstream_density(cs.DM(rho_firsts)).full().squeeze(),
        )

    def test_links__get_flow(self):
        N = self.N
        rho = np.maximum(0, np.random.randn(N) * 5 + 30)
        v = np.maximum(0, np.random.randn(N) * 10 + 80)
        lanes = np.random.randint(3, 7)
        args = [rho, v, lanes]
        np.testing.assert_allclose(
            self.NE.links.get_flow(*args),
            self.CE.links.get_flow(*map(cs.DM, args)).full().squeeze(),
        )

    def test_links__step_density(self):
        N = self.N
        rho = np.maximum(0, np.random.randn(N) * 5 + 30)
        q = np.maximum(0, np.random.randn(N) * 10 + 800)
        q_up = np.maximum(0, np.random.randn(N) * 10 + 800)
        lanes = np.random.randint(3, 7)
        L = np.random.rand() * 1000 + 1000
        T = np.random.rand()
        args = [rho, q, q_up, lanes, L, T]
        np.testing.assert_allclose(
            self.NE.links.step_density(*args),
            self.CE.links.step_density(*map(cs.DM, args)).full().squeeze(),
        )

    def test_links__step_speed(self):
        N = self.N
        v = np.maximum(0, np.random.randn(N) * 10 + 100)
        v_up = np.maximum(0, np.random.randn(N) * 10 + 100)
        rho = np.maximum(0, np.random.randn(N) * 5 + 30)
        rho_down = np.maximum(0, np.random.randn(N) * 5 + 30)
        Veq = np.maximum(0, np.random.randn(N) * 10 + 100)
        lanes = np.random.randint(3, 7)
        L = np.random.rand() * 1000 + 1000
        tau = np.random.rand()
        eta = T = np.random.rand()
        kappa = np.random.rand()
        T = np.random.rand()
        q_ramp = np.maximum(0, np.random.randn() * 100 + 800)
        delta = np.random.rand()
        lanes_drop = np.random.randint(1, 3)
        phi = np.random.rand()
        rho_crit = np.random.rand() * 10 + 30
        args = [
            v,
            v_up,
            rho,
            rho_down,
            Veq,
            lanes,
            L,
            tau,
            eta,
            kappa,
            T,
            q_ramp,
            delta,
            lanes_drop,
            phi,
            rho_crit,
        ]
        np.testing.assert_allclose(
            self.NE.links.step_speed(*args),
            self.CE.links.step_speed(*map(cs.DM, args)).full().squeeze(),
        )

    def test_links__Veq(self):
        N = self.N
        rho = np.maximum(0, np.random.randn(N) * 5 + 30)
        v_free = np.random.rand() * 40 + 80
        rho_crit = np.random.rand() * 10 + 30
        a = np.random.rand() + 1
        args = [rho, v_free, rho_crit, a]
        np.testing.assert_allclose(
            self.NE.links.Veq(*args),
            self.CE.links.Veq(*map(cs.DM, args)).full().squeeze(),
        )

    def test_origins__step_queue(self):
        w = np.random.randint(low=1, high=3) * 100 + 300
        d = np.random.randint(low=1, high=3) * 100 + 300
        q = np.maximum(0, np.random.randn() * 200 + 800)
        T = np.random.rand()
        args = [w, d, q, T]
        np.testing.assert_allclose(
            self.NE.origins.step_queue(*args),
            self.CE.origins.step_queue(*map(cs.DM, args)).full().squeeze(),
        )

    def test_origins__get_ramp_flow(self):
        w = np.random.randint(low=1, high=3) * 100 + 300
        d = np.random.randint(low=1, high=3) * 100 + 300
        C = np.maximum(100, np.random.randn() * 500 + 2000)
        r = np.random.rand()
        rho_max = (np.random.rand() * 30 + 130) * 2
        rho_first = np.random.rand() * 20 + 20
        rho_crit = np.random.rand() * 20 + 20
        T = np.random.rand()
        args = [d, w, C, r, rho_max, rho_first, rho_crit, T]
        for type_ in ("in", "out"):
            np.testing.assert_allclose(
                self.NE.origins.get_ramp_flow(*args, type_),
                self.CE.origins.get_ramp_flow(*map(cs.DM, args), type_)
                .full()
                .squeeze(),
            )

    def test_destinations__get_ramp_flow(self):
        rho_last = np.random.rand() * 20 + 20
        rho_destination = np.random.rand() * 30 + 20
        rho_crit = np.random.rand() * 20 + 20
        args = [rho_last, rho_destination, rho_crit]
        np.testing.assert_allclose(
            self.NE.destinations.get_congested_downstream_density(*args),
            self.CE.destinations.get_congested_downstream_density(*map(cs.DM, args))
            .full()
            .squeeze(),
        )


if __name__ == "__main__":
    unittest.main()



import re
import unittest
from typing import Iterable

from sym_metanet import Destination, Link, MeteredOnRamp, Network, Node, Origin, engines

engine = engines.use("numpy", var_type="randn")


class TestNetwork(unittest.TestCase):
    def assertAtLeastOneRegexMatch(self, pattern: str, strings: Iterable[str]):
        for string in strings:
            if re.match(pattern, string):
                return
        raise self.failureException(f"no string matches pattern: {pattern}")

    def test_add_node__adds_node_correctly(self):
        node = Node(name="This is a random name")
        net = Network(name="Another random name")
        net.add_node(node)
        self.assertIn(node, net.nodes)
        self.assertIn(node.name, net.nodes_by_name)

    def test_add_nodes__adds_nodes_correctly(self):
        node1 = Node(name="This is a random name1")
        node2 = Node(name="This is a random name2")
        net = Network(name="Another random name")
        net.add_nodes((node1, node2))
        self.assertIn(node1, net.nodes)
        self.assertIn(node2, net.nodes)
        self.assertIn(node1.name, net.nodes_by_name)
        self.assertIn(node1.name, net.nodes_by_name)

    def test_add_link__adds_link_correctly(self):
        upnode = Node(name="N1")
        downnode = Node(name="N2")
        link = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        net = Network(name="Net")
        net.add_nodes((upnode, downnode))
        net.add_link(upnode, link, downnode)
        self.assertIs(link, net.links[upnode, downnode])
        self.assertEqual(len(net.links), 1)
        self.assertIn(link.name, net.links_by_name)
        self.assertIn(link, net.nodes_by_link)
        self.assertEqual(net.nodes_by_link[link], (upnode, downnode))

    def test_add_links__adds_links_correctly(self):
        upnode1 = Node(name="N11")
        downnode1 = Node(name="N12")
        link1 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        upnode2 = Node(name="N21")
        downnode2 = Node(name="N21")
        link2 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L2")
        net = Network(name="Net")
        net.add_nodes((upnode1, downnode1, upnode2, downnode2))
        net.add_links(((upnode1, link1, downnode1), (upnode2, link2, downnode2)))
        self.assertEqual(len(net.links), 2)
        for data in ((link1, upnode1, downnode1), (link2, upnode2, downnode2)):
            link, nodeup, nodedown = data
            self.assertEqual(net.nodes_by_link[link], (nodeup, nodedown))
            self.assertIs(link, net.links[nodeup, nodedown])
            self.assertIn(link.name, net.links_by_name)
            self.assertIn(link, net.nodes_by_link)

    def test_add_path__adds_nodes_and_links_correctly(self):
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        N3 = Node(name="N3")
        L1 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        L2 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L2")
        L3 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L3")
        O1 = Origin(name="23423")
        D1 = Destination(name="23421")
        net = Network(name="Net")
        net.add_path(origin=O1, path=(N1, L1, N2), destination=D1)
        net.add_path(origin=O1, path=(N1, L2, N3, L3, N2), destination=D1)
        for n in (N1, N2, N3):
            self.assertIn(n, net.nodes)
        self.assertEqual(len(net.links), 3)
        self.assertIs(L1, net.links[N1, N2])
        self.assertIs(L2, net.links[N1, N3])
        self.assertIs(L3, net.links[N3, N2])
        self.assertIn(O1, net.origins)
        self.assertIn(D1, net.destinations)

    def test_add_path__raises__with_single_node(self):
        net = Network(name="Net")
        with self.assertRaises(ValueError):
            net.add_path(path=(Node(name="N1"),))

    def test_add_origin__adds_origin_correctly(self):
        node = Node(name="This is a random name")
        origin = Origin(name="23423")
        net = Network(name="Another random name")
        net.add_node(node)
        net.add_origin(origin, node)
        self.assertIn(origin, net.origins)
        self.assertIs(net.origins[origin], node)
        self.assertIn(origin.name, net.origins_by_name)
        self.assertIs(net.origins_by_name[origin.name], origin)
        self.assertIn(node, net.origins_by_node)
        self.assertIs(net.origins_by_node[node], origin)

    def test_add_destination__adds_destination_correctly(self):
        node = Node(name="This is a random name")
        destination = Destination(name="23423")
        net = Network(name="Another random name")
        net.add_node(node)
        net.add_destination(destination, node)
        self.assertIn(destination, net.destinations)
        self.assertIs(net.destinations[destination], node)
        self.assertIn(destination.name, net.destinations_by_name)
        self.assertIs(net.destinations_by_name[destination.name], destination)
        self.assertIn(node, net.destinations_by_node)
        self.assertIs(net.destinations_by_node[node], destination)

    def test_out_links__gets_correct_outward_links(self):
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        N3 = Node(name="N3")
        L1 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        L2 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L2")
        net = Network(name=".Net")
        net.add_links(((N1, L1, N2), (N1, L2, N3)))
        self.assertEqual({(N1, N3, L2), (N1, N2, L1)}, set(net.out_links(N1)))

    def test_in_links__gets_correct_inward_links(self):
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        N3 = Node(name="N3")
        L1 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        L2 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L2")
        net = Network(name=".Net")
        net.add_links(((N1, L1, N2), (N3, L2, N2)))
        self.assertEqual({(N1, N2, L1), (N3, N2, L2)}, set(net.in_links(N2)))

    def test_is_valid__raises__condition_1a(self):
        # duplicate link
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        N3 = Node(name="N3")
        L = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        net = Network(name=".Net")
        net.add_nodes((N1, N2, N3))
        net.add_link(N1, L, N2)
        net.add_link(N1, L, N3)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch(
            "Element L1 is duplicated in the network.", msgs
        )

    def test_is_valid__raises__condition_1b(self):
        # duplicate origin
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        net = Network(name=".Net")
        O = MeteredOnRamp(100, name="O")
        net.add_origin(O, N1)
        net.add_origin(O, N2)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch("Element O is duplicated in the network.", msgs)

    def test_is_valid__raises__condition_1c(self):
        # duplicate destination
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        net = Network(name=".Net")
        D = Destination(name="D")
        net.add_destination(D, N1)
        net.add_destination(D, N2)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch("Element D is duplicated in the network.", msgs)

    def test_is_valid__raises__condition2(self):
        N = Node(name="N")
        O = MeteredOnRamp(100, name="23423")
        D = Destination(name="23423")
        net = Network(name=".Net")
        net.add_node(N)
        net.add_origin(O, N)
        net.add_destination(D, N)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch(
            "Node N must either have an origin or a destination, but not " "both.", msgs
        )

    def test_is_valid__raises__condition_3(self):
        net = Network(name=".Net")
        net.add_node(Node(name="N"))
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch("Node N is connected to no link.", msgs)

    def test_is_valid__raises__condition_4_and_5(self):
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        L = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        net = Network(name=".Net that does not raise")
        net.add_link(N1, L, N2)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch(
            "Node N1 has neither any entering links nor an origin.", msgs
        )
        self.assertAtLeastOneRegexMatch(
            "Node N2 has neither any exiting links nor a destination.", msgs
        )

    def test_is_valid__raises__condition_6(self):
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        N3 = Node(name="N3")
        L1 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        L2 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L3")
        O = Origin(name="O")
        net = Network(name=".Net")
        net.add_nodes((N1, N2, N3))
        net.add_link(N1, L1, N2)
        net.add_link(N2, L2, N3)
        net.add_origin(O, N2)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch(
            r"Expected node N2 to have no entering links, as it is connected "
            r"to origin O \(only ramps support entering links\).",
            msgs,
        )

    def test_is_valid__raises__condition_7(self):
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        N3 = Node(name="N3")
        L1 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        L2 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L3")
        O = Origin(name="O")
        net = Network(name=".Net")
        net.add_nodes((N1, N2, N3))
        net.add_link(N1, L1, N2)
        net.add_link(N1, L2, N3)
        net.add_origin(O, N1)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch(
            "Expected node N1 to have at most one exiting link, as it is "
            "connected to origin O.",
            msgs,
        )

    def test_is_valid__raises__condition_8(self):
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        N3 = Node(name="N3")
        L1 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        L2 = Link(4, 3, 1, 180, 30, 100, 1.8, name="L3")
        D = Destination(name="D")
        net = Network(name=".Net")
        net.add_nodes((N1, N2, N3))
        net.add_link(N1, L1, N3)
        net.add_link(N2, L2, N3)
        net.add_destination(D, N3)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch(
            "Expected node N3 to have at most one entering link, as it is "
            "connected to destination D.",
            msgs,
        )

    def test_is_valid__raises__condition_9(self):
        N1 = Node(name="N1")
        N2 = Node(name="N2")
        L = Link(4, 3, 1, 180, 30, 100, 1.8, name="L1")
        net = Network(name=".Net that does not raise")
        D = Destination(name="D")
        net.add_link(N1, L, N2)
        net.add_destination(D, N1)
        ok, msgs = net.is_valid(raises=False)
        self.assertFalse(ok)
        self.assertAtLeastOneRegexMatch(
            "Expected node N1 to have no exiting links, as it is connected "
            "to destination D.",
            msgs,
        )


if __name__ == "__main__":
    unittest.main()
